# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

# For no Pydantic environment, we need to skip the tests
pytest.importorskip("google.cloud.aiplatform_v1")

from google.api_core.gapic_v1.method import DEFAULT

from airflow.providers.google.cloud.hooks.vertex_ai.prediction_service import (
    PredictionServiceHook,
)

from unit.google.cloud.utils.base_gcp_mock import (
    mock_base_gcp_hook_default_project_id,
    mock_base_gcp_hook_no_default_project_id,
)

TEST_GCP_CONN_ID: str = "test-gcp-conn-id"
TEST_REGION: str = "test-region"
TEST_PROJECT_ID: str = "test-project-id"
TEST_ENDPOINT_ID: str = "test-endpoint-id"
TEST_OUTPUT_CONFIG: dict = {}

BASE_STRING = "airflow.providers.google.common.hooks.base_google.{}"
PREDICTION_SERVICE_STRING = "airflow.providers.google.cloud.hooks.vertex_ai.prediction_service.{}"


class TestPredictionServiceWithDefaultProjectIdHook:
    def setup_method(self):
        with mock.patch(
            BASE_STRING.format("GoogleBaseHook.__init__"), new=mock_base_gcp_hook_default_project_id
        ):
            self.hook = PredictionServiceHook(gcp_conn_id=TEST_GCP_CONN_ID)

    @mock.patch(PREDICTION_SERVICE_STRING.format("PredictionServiceHook.get_prediction_service_client"))
    def test_predict(self, mock_client):
        self.hook.predict(
            endpoint_id=TEST_ENDPOINT_ID,
            instances=["instance1", "instance2"],
            project_id=TEST_PROJECT_ID,
            location=TEST_REGION,
        )
        mock_client.assert_called_once_with(TEST_REGION)
        mock_client.return_value.predict.assert_called_once_with(
            request=dict(
                endpoint=f"projects/{TEST_PROJECT_ID}/locations/{TEST_REGION}/endpoints/{TEST_ENDPOINT_ID}",
                instances=["instance1", "instance2"],
                parameters=None,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )


class TestPredictionServiceWithoutDefaultProjectIdHook:
    def setup_method(self):
        with mock.patch(
            BASE_STRING.format("GoogleBaseHook.__init__"), new=mock_base_gcp_hook_no_default_project_id
        ):
            self.hook = PredictionServiceHook(gcp_conn_id=TEST_GCP_CONN_ID)

    @mock.patch(PREDICTION_SERVICE_STRING.format("PredictionServiceHook.get_prediction_service_client"))
    def test_predict(self, mock_client):
        self.hook.predict(
            endpoint_id=TEST_ENDPOINT_ID,
            instances=["instance1", "instance2"],
            project_id=TEST_PROJECT_ID,
            location=TEST_REGION,
        )
        mock_client.assert_called_once_with(TEST_REGION)
        mock_client.return_value.predict.assert_called_once_with(
            request=dict(
                endpoint=f"projects/{TEST_PROJECT_ID}/locations/{TEST_REGION}/endpoints/{TEST_ENDPOINT_ID}",
                instances=["instance1", "instance2"],
                parameters=None,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

from airflow.providers.google.cloud.transfers.adls_to_gcs import ADLSToGCSOperator

TASK_ID = "test-adls-gcs-operator"
ADLS_PATH_1 = "*"
GCS_PATH = "gs://test/"
MOCK_FILES = [
    "test/TEST1.csv",
    "test/TEST2.csv",
    "test/path/TEST3.csv",
    "test/path/PARQUET.parquet",
    "test/path/PIC.png",
]
AZURE_CONN_ID = "azure_data_lake_default"
GCS_CONN_ID = "google_cloud_default"
IMPERSONATION_CHAIN = ["ACCOUNT_1", "ACCOUNT_2", "ACCOUNT_3"]


class TestAdlsToGoogleCloudStorageOperator:
    def test_init(self):
        """Test AdlsToGoogleCloudStorageOperator instance is properly initialized."""

        operator = ADLSToGCSOperator(
            task_id=TASK_ID,
            src_adls=ADLS_PATH_1,
            dest_gcs=GCS_PATH,
            replace=False,
            azure_data_lake_conn_id=AZURE_CONN_ID,
            gcp_conn_id=GCS_CONN_ID,
        )

        assert operator.task_id == TASK_ID
        assert operator.src_adls == ADLS_PATH_1
        assert operator.dest_gcs == GCS_PATH
        assert operator.replace is False
        assert operator.gcp_conn_id == GCS_CONN_ID
        assert operator.azure_data_lake_conn_id == AZURE_CONN_ID

    @mock.patch("airflow.providers.google.cloud.transfers.adls_to_gcs.AzureDataLakeHook")
    @mock.patch("airflow.providers.microsoft.azure.operators.adls.AzureDataLakeHook")
    @mock.patch("airflow.providers.google.cloud.transfers.adls_to_gcs.GCSHook")
    def test_execute(self, gcs_mock_hook, adls_one_mock_hook, adls_two_mock_hook):
        """Test the execute function when the run is successful."""

        operator = ADLSToGCSOperator(
            task_id=TASK_ID,
            src_adls=ADLS_PATH_1,
            dest_gcs=GCS_PATH,
            replace=False,
            azure_data_lake_conn_id=AZURE_CONN_ID,
            gcp_conn_id=GCS_CONN_ID,
            google_impersonation_chain=IMPERSONATION_CHAIN,
        )

        adls_one_mock_hook.return_value.list.return_value = MOCK_FILES
        adls_two_mock_hook.return_value.list.return_value = MOCK_FILES

        # gcs_mock_hook.return_value.upload.side_effect = _assert_upload
        uploaded_files = operator.execute(None)
        gcs_mock_hook.return_value.upload.assert_has_calls(
            [
                mock.call(
                    bucket_name="test", filename=mock.ANY, object_name="test/path/PARQUET.parquet", gzip=False
                ),
                mock.call(
                    bucket_name="test", filename=mock.ANY, object_name="test/path/TEST3.csv", gzip=False
                ),
                mock.call(bucket_name="test", filename=mock.ANY, object_name="test/path/PIC.png", gzip=False),
                mock.call(bucket_name="test", filename=mock.ANY, object_name="test/TEST1.csv", gzip=False),
                mock.call(bucket_name="test", filename=mock.ANY, object_name="test/TEST2.csv", gzip=False),
            ],
            any_order=True,
        )

        adls_one_mock_hook.assert_called_once_with(azure_data_lake_conn_id=AZURE_CONN_ID)
        adls_two_mock_hook.assert_called_once_with(azure_data_lake_conn_id=AZURE_CONN_ID)
        gcs_mock_hook.assert_called_once_with(
            gcp_conn_id=GCS_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        # we expect MOCK_FILES to be uploaded
        assert sorted(MOCK_FILES) == sorted(uploaded_files)

    @mock.patch("airflow.providers.google.cloud.transfers.adls_to_gcs.AzureDataLakeHook")
    @mock.patch("airflow.providers.microsoft.azure.operators.adls.AzureDataLakeHook")
    @mock.patch("airflow.providers.google.cloud.transfers.adls_to_gcs.GCSHook")
    def test_execute_with_gzip(self, gcs_mock_hook, adls_one_mock_hook, adls_two_mock_hook):
        """Test the execute function when the run is successful."""

        operator = ADLSToGCSOperator(
            task_id=TASK_ID,
            src_adls=ADLS_PATH_1,
            dest_gcs=GCS_PATH,
            replace=False,
            azure_data_lake_conn_id=AZURE_CONN_ID,
            gcp_conn_id=GCS_CONN_ID,
            gzip=True,
        )

        adls_one_mock_hook.return_value.list.return_value = MOCK_FILES
        adls_two_mock_hook.return_value.list.return_value = MOCK_FILES

        # gcs_mock_hook.return_value.upload.side_effect = _assert_upload
        uploaded_files = operator.execute(None)
        gcs_mock_hook.return_value.upload.assert_has_calls(
            [
                mock.call(
                    bucket_name="test", filename=mock.ANY, object_name="test/path/PARQUET.parquet", gzip=True
                ),
                mock.call(
                    bucket_name="test", filename=mock.ANY, object_name="test/path/TEST3.csv", gzip=True
                ),
                mock.call(bucket_name="test", filename=mock.ANY, object_name="test/path/PIC.png", gzip=True),
                mock.call(bucket_name="test", filename=mock.ANY, object_name="test/TEST1.csv", gzip=True),
                mock.call(bucket_name="test", filename=mock.ANY, object_name="test/TEST2.csv", gzip=True),
            ],
            any_order=True,
        )

        # we expect MOCK_FILES to be uploaded
        assert sorted(MOCK_FILES) == sorted(uploaded_files)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import PropertyMock

import google
from google.cloud.bigtable import Client, enums
from google.cloud.bigtable.instance import Instance

from airflow.providers.google.cloud.hooks.bigtable import BigtableHook
from airflow.providers.google.common.consts import CLIENT_INFO

from unit.google.cloud.utils.base_gcp_mock import (
    GCP_PROJECT_ID_HOOK_UNIT_TEST,
    mock_base_gcp_hook_default_project_id,
    mock_base_gcp_hook_no_default_project_id,
)

CBT_INSTANCE = "instance"
CBT_INSTANCE_DISPLAY_NAME = "test instance"
CBT_INSTANCE_TYPE = enums.Instance.Type.PRODUCTION
CBT_INSTANCE_LABELS = {"env": "sit"}
CBT_CLUSTER = "cluster"
CBT_ZONE = "zone"
CBT_TABLE = "table"
CBT_REPLICA_CLUSTER_ID = "replica-cluster"
CBT_REPLICA_CLUSTER_ZONE = "us-west1-b"
CBT_REPLICATE_CLUSTERS = [
    {"id": "replica-1", "zone": "us-west1-a"},
    {"id": "replica-2", "zone": "us-central1-f"},
    {"id": "replica-3", "zone": "us-east1-d"},
]


class TestBigtableHookNoDefaultProjectId:
    def setup_method(self):
        with mock.patch(
            "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
            new=mock_base_gcp_hook_no_default_project_id,
        ):
            self.bigtable_hook_no_default_project_id = BigtableHook(gcp_conn_id="test")

    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook.get_credentials")
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.Client")
    def test_bigtable_client_creation(self, mock_client, mock_get_creds):
        result = self.bigtable_hook_no_default_project_id._get_client(GCP_PROJECT_ID_HOOK_UNIT_TEST)
        mock_client.assert_called_once_with(
            project=GCP_PROJECT_ID_HOOK_UNIT_TEST,
            credentials=mock_get_creds.return_value,
            client_info=CLIENT_INFO,
            admin=True,
        )
        assert mock_client.return_value == result
        assert self.bigtable_hook_no_default_project_id._client == result

    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_get_instance_overridden_project_id(self, get_client):
        instance_method = get_client.return_value.instance
        instance_exists_method = instance_method.return_value.exists
        instance_exists_method.return_value = True
        res = self.bigtable_hook_no_default_project_id.get_instance(
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST, instance_id=CBT_INSTANCE
        )
        instance_method.assert_called_once_with("instance")
        instance_exists_method.assert_called_once_with()
        get_client.assert_called_once_with(project_id="example-project")
        assert res is not None

    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_delete_instance_overridden_project_id(self, get_client):
        instance_method = get_client.return_value.instance
        instance_exists_method = instance_method.return_value.exists
        instance_exists_method.return_value = True
        delete_method = instance_method.return_value.delete
        res = self.bigtable_hook_no_default_project_id.delete_instance(
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST, instance_id=CBT_INSTANCE
        )
        instance_method.assert_called_once_with("instance")
        instance_exists_method.assert_called_once_with()
        delete_method.assert_called_once_with()
        get_client.assert_called_once_with(project_id="example-project")
        assert res is None

    @mock.patch("google.cloud.bigtable.instance.Instance.create")
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_create_instance_overridden_project_id(self, get_client, instance_create):
        operation = mock.Mock()
        operation.result_return_value = Instance(instance_id=CBT_INSTANCE, client=get_client)
        instance_create.return_value = operation
        res = self.bigtable_hook_no_default_project_id.create_instance(
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
            instance_id=CBT_INSTANCE,
            main_cluster_id=CBT_CLUSTER,
            main_cluster_zone=CBT_ZONE,
        )
        get_client.assert_called_once_with(project_id="example-project")
        instance_create.assert_called_once_with(clusters=mock.ANY)
        assert res.instance_id == "instance"

    @mock.patch("google.cloud.bigtable.instance.Instance.update")
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_update_instance_overridden_project_id(self, get_client, instance_update):
        operation = mock.Mock()
        operation.result_return_value = Instance(instance_id=CBT_INSTANCE, client=get_client)
        instance_update.return_value = operation
        res = self.bigtable_hook_no_default_project_id.update_instance(
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
            instance_id=CBT_INSTANCE,
            instance_display_name=CBT_INSTANCE_DISPLAY_NAME,
            instance_type=CBT_INSTANCE_TYPE,
            instance_labels=CBT_INSTANCE_LABELS,
        )
        get_client.assert_called_once_with(project_id="example-project")
        instance_update.assert_called_once_with()
        assert res.instance_id == "instance"

    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_delete_table_overridden_project_id(self, get_client):
        instance_method = get_client.return_value.instance
        instance_exists_method = instance_method.return_value.exists
        table_delete_method = instance_method.return_value.table.return_value.delete
        instance_exists_method.return_value = True
        self.bigtable_hook_no_default_project_id.delete_table(
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST, instance_id=CBT_INSTANCE, table_id=CBT_TABLE
        )
        get_client.assert_called_once_with(project_id="example-project")
        instance_exists_method.assert_called_once_with()
        table_delete_method.assert_called_once_with()


class TestBigtableHookDefaultProjectId:
    def setup_method(self):
        with mock.patch(
            "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
            new=mock_base_gcp_hook_default_project_id,
        ):
            self.bigtable_hook_default_project_id = BigtableHook(gcp_conn_id="test")

    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook.get_credentials")
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.Client")
    def test_bigtable_client_creation(self, mock_client, mock_get_creds):
        result = self.bigtable_hook_default_project_id._get_client(GCP_PROJECT_ID_HOOK_UNIT_TEST)
        mock_client.assert_called_once_with(
            project=GCP_PROJECT_ID_HOOK_UNIT_TEST,
            credentials=mock_get_creds.return_value,
            client_info=CLIENT_INFO,
            admin=True,
        )
        assert mock_client.return_value == result
        assert self.bigtable_hook_default_project_id._client == result

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_get_instance(self, get_client, mock_project_id):
        instance_method = get_client.return_value.instance
        instance_exists_method = instance_method.return_value.exists
        instance_exists_method.return_value = True
        res = self.bigtable_hook_default_project_id.get_instance(
            instance_id=CBT_INSTANCE,
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
        )
        instance_method.assert_called_once_with("instance")
        instance_exists_method.assert_called_once_with()
        get_client.assert_called_once_with(project_id="example-project")
        assert res is not None

    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_get_instance_overridden_project_id(self, get_client):
        instance_method = get_client.return_value.instance
        instance_exists_method = instance_method.return_value.exists
        instance_exists_method.return_value = True
        res = self.bigtable_hook_default_project_id.get_instance(
            project_id="new-project", instance_id=CBT_INSTANCE
        )
        instance_method.assert_called_once_with("instance")
        instance_exists_method.assert_called_once_with()
        get_client.assert_called_once_with(project_id="new-project")
        assert res is not None

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_get_instance_no_instance(self, get_client, mock_project_id):
        instance_method = get_client.return_value.instance
        instance_exists_method = instance_method.return_value.exists
        instance_exists_method.return_value = False
        res = self.bigtable_hook_default_project_id.get_instance(
            instance_id=CBT_INSTANCE,
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
        )
        instance_method.assert_called_once_with("instance")
        instance_exists_method.assert_called_once_with()
        get_client.assert_called_once_with(project_id="example-project")
        assert res is None

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_delete_instance(self, get_client, mock_project_id):
        instance_method = get_client.return_value.instance
        instance_exists_method = instance_method.return_value.exists
        instance_exists_method.return_value = True
        delete_method = instance_method.return_value.delete
        res = self.bigtable_hook_default_project_id.delete_instance(
            instance_id=CBT_INSTANCE,
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
        )
        instance_method.assert_called_once_with("instance")
        instance_exists_method.assert_called_once_with()
        delete_method.assert_called_once_with()
        get_client.assert_called_once_with(project_id="example-project")
        assert res is None

    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_delete_instance_overridden_project_id(self, get_client):
        instance_method = get_client.return_value.instance
        instance_exists_method = instance_method.return_value.exists
        instance_exists_method.return_value = True
        delete_method = instance_method.return_value.delete
        res = self.bigtable_hook_default_project_id.delete_instance(
            project_id="new-project", instance_id=CBT_INSTANCE
        )
        instance_method.assert_called_once_with("instance")
        instance_exists_method.assert_called_once_with()
        delete_method.assert_called_once_with()
        get_client.assert_called_once_with(project_id="new-project")
        assert res is None

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_delete_instance_no_instance(self, get_client, mock_project_id):
        instance_method = get_client.return_value.instance
        instance_exists_method = instance_method.return_value.exists
        instance_exists_method.return_value = False
        delete_method = instance_method.return_value.delete
        self.bigtable_hook_default_project_id.delete_instance(
            instance_id=CBT_INSTANCE,
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
        )
        instance_method.assert_called_once_with("instance")
        instance_exists_method.assert_called_once_with()
        delete_method.assert_not_called()
        get_client.assert_called_once_with(project_id="example-project")

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("google.cloud.bigtable.instance.Instance.create")
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_create_instance(self, get_client, instance_create, mock_project_id):
        operation = mock.Mock()
        operation.result_return_value = Instance(instance_id=CBT_INSTANCE, client=get_client)
        instance_create.return_value = operation
        res = self.bigtable_hook_default_project_id.create_instance(
            instance_id=CBT_INSTANCE,
            main_cluster_id=CBT_CLUSTER,
            main_cluster_zone=CBT_ZONE,
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
        )
        get_client.assert_called_once_with(project_id="example-project")
        instance_create.assert_called_once_with(clusters=mock.ANY)
        assert res.instance_id == "instance"

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("google.cloud.bigtable.instance.Instance.cluster")
    @mock.patch("google.cloud.bigtable.instance.Instance.create")
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_create_instance_with_one_replica_cluster_production(
        self, get_client, instance_create, cluster, mock_project_id
    ):
        operation = mock.Mock()
        operation.result_return_value = Instance(instance_id=CBT_INSTANCE, client=get_client)
        instance_create.return_value = operation

        res = self.bigtable_hook_default_project_id.create_instance(
            instance_id=CBT_INSTANCE,
            main_cluster_id=CBT_CLUSTER,
            main_cluster_zone=CBT_ZONE,
            replica_clusters=[{"id": CBT_REPLICA_CLUSTER_ID, "zone": CBT_REPLICA_CLUSTER_ZONE}],
            cluster_nodes=1,
            cluster_storage_type=enums.StorageType.SSD,
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
            instance_type=enums.Instance.Type.PRODUCTION,
        )
        cluster.assert_has_calls(
            [
                mock.call(
                    cluster_id=CBT_CLUSTER,
                    location_id=CBT_ZONE,
                    serve_nodes=1,
                    default_storage_type=enums.StorageType.SSD,
                ),
                mock.call(CBT_REPLICA_CLUSTER_ID, CBT_REPLICA_CLUSTER_ZONE, 1, enums.StorageType.SSD),
            ],
            any_order=True,
        )
        get_client.assert_called_once_with(project_id="example-project")
        instance_create.assert_called_once_with(clusters=mock.ANY)
        assert res.instance_id == "instance"

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("google.cloud.bigtable.instance.Instance.cluster")
    @mock.patch("google.cloud.bigtable.instance.Instance.create")
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_create_instance_with_one_replica_cluster_development(
        self, get_client, instance_create, cluster, mock_project_id
    ):
        operation = mock.Mock()
        operation.result_return_value = Instance(instance_id=CBT_INSTANCE, client=get_client)
        instance_create.return_value = operation

        res = self.bigtable_hook_default_project_id.create_instance(
            instance_id=CBT_INSTANCE,
            main_cluster_id=CBT_CLUSTER,
            main_cluster_zone=CBT_ZONE,
            replica_clusters=[{"id": CBT_REPLICA_CLUSTER_ID, "zone": CBT_REPLICA_CLUSTER_ZONE}],
            cluster_nodes=1,
            cluster_storage_type=enums.StorageType.SSD,
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
            instance_type=enums.Instance.Type.DEVELOPMENT,
        )
        cluster.assert_has_calls(
            [
                mock.call(
                    cluster_id=CBT_CLUSTER, location_id=CBT_ZONE, default_storage_type=enums.StorageType.SSD
                ),
                mock.call(CBT_REPLICA_CLUSTER_ID, CBT_REPLICA_CLUSTER_ZONE, 1, enums.StorageType.SSD),
            ],
            any_order=True,
        )
        get_client.assert_called_once_with(project_id="example-project")
        instance_create.assert_called_once_with(clusters=mock.ANY)
        assert res.instance_id == "instance"

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("google.cloud.bigtable.instance.Instance.cluster")
    @mock.patch("google.cloud.bigtable.instance.Instance.create")
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_create_instance_with_multiple_replica_clusters(
        self, get_client, instance_create, cluster, mock_project_id
    ):
        operation = mock.Mock()
        operation.result_return_value = Instance(instance_id=CBT_INSTANCE, client=get_client)
        instance_create.return_value = operation

        res = self.bigtable_hook_default_project_id.create_instance(
            instance_id=CBT_INSTANCE,
            main_cluster_id=CBT_CLUSTER,
            main_cluster_zone=CBT_ZONE,
            replica_clusters=CBT_REPLICATE_CLUSTERS,
            cluster_nodes=1,
            cluster_storage_type=enums.StorageType.SSD,
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
        )
        cluster.assert_has_calls(
            [
                mock.call(
                    cluster_id=CBT_CLUSTER,
                    location_id=CBT_ZONE,
                    serve_nodes=1,
                    default_storage_type=enums.StorageType.SSD,
                ),
                mock.call("replica-1", "us-west1-a", 1, enums.StorageType.SSD),
                mock.call("replica-2", "us-central1-f", 1, enums.StorageType.SSD),
                mock.call("replica-3", "us-east1-d", 1, enums.StorageType.SSD),
            ],
            any_order=True,
        )
        get_client.assert_called_once_with(project_id="example-project")
        instance_create.assert_called_once_with(clusters=mock.ANY)
        assert res.instance_id == "instance"

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("google.cloud.bigtable.instance.Instance.update")
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_update_instance(self, get_client, instance_update, mock_project_id):
        operation = mock.Mock()
        operation.result_return_value = Instance(instance_id=CBT_INSTANCE, client=get_client)
        instance_update.return_value = operation
        res = self.bigtable_hook_default_project_id.update_instance(
            instance_id=CBT_INSTANCE,
            instance_display_name=CBT_INSTANCE_DISPLAY_NAME,
            instance_type=CBT_INSTANCE_TYPE,
            instance_labels=CBT_INSTANCE_LABELS,
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
        )
        get_client.assert_called_once_with(project_id="example-project")
        instance_update.assert_called_once_with()
        assert res.instance_id == "instance"

    @mock.patch("google.cloud.bigtable.instance.Instance.create")
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_create_instance_overridden_project_id(self, get_client, instance_create):
        operation = mock.Mock()
        operation.result_return_value = Instance(instance_id=CBT_INSTANCE, client=get_client)
        instance_create.return_value = operation
        res = self.bigtable_hook_default_project_id.create_instance(
            project_id="new-project",
            instance_id=CBT_INSTANCE,
            main_cluster_id=CBT_CLUSTER,
            main_cluster_zone=CBT_ZONE,
        )
        get_client.assert_called_once_with(project_id="new-project")
        instance_create.assert_called_once_with(clusters=mock.ANY)
        assert res.instance_id == "instance"

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_delete_table(self, get_client, mock_project_id):
        instance_method = get_client.return_value.instance
        instance_exists_method = instance_method.return_value.exists
        table_delete_method = instance_method.return_value.table.return_value.delete
        instance_exists_method.return_value = True
        self.bigtable_hook_default_project_id.delete_table(
            instance_id=CBT_INSTANCE,
            table_id=CBT_TABLE,
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
        )
        get_client.assert_called_once_with(project_id="example-project")
        instance_exists_method.assert_called_once_with()
        table_delete_method.assert_called_once_with()

    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_delete_table_overridden_project_id(self, get_client):
        instance_method = get_client.return_value.instance
        instance_exists_method = instance_method.return_value.exists
        table_delete_method = instance_method.return_value.table.return_value.delete
        instance_exists_method.return_value = True
        self.bigtable_hook_default_project_id.delete_table(
            project_id="new-project", instance_id=CBT_INSTANCE, table_id=CBT_TABLE
        )
        get_client.assert_called_once_with(project_id="new-project")
        instance_exists_method.assert_called_once_with()
        table_delete_method.assert_called_once_with()

    @mock.patch("google.cloud.bigtable.table.Table.create")
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_create_table(self, get_client, create):
        instance_method = get_client.return_value.instance
        instance_exists_method = instance_method.return_value.exists
        instance_exists_method.return_value = True
        client = mock.Mock(Client)
        instance = google.cloud.bigtable.instance.Instance(instance_id=CBT_INSTANCE, client=client)
        self.bigtable_hook_default_project_id.create_table(instance=instance, table_id=CBT_TABLE)
        get_client.assert_not_called()
        create.assert_called_once_with([], {})

    @mock.patch("google.cloud.bigtable.cluster.Cluster.update")
    @mock.patch("google.cloud.bigtable.cluster.Cluster.reload")
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_update_cluster(self, get_client, reload, update):
        instance_method = get_client.return_value.instance
        instance_exists_method = instance_method.return_value.exists
        instance_exists_method.return_value = True
        client = mock.Mock(Client)
        instance = google.cloud.bigtable.instance.Instance(instance_id=CBT_INSTANCE, client=client)
        self.bigtable_hook_default_project_id.update_cluster(
            instance=instance, cluster_id=CBT_CLUSTER, nodes=4
        )
        get_client.assert_not_called()
        reload.assert_called_once_with()
        update.assert_called_once_with()

    @mock.patch("google.cloud.bigtable.table.Table.list_column_families")
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_list_column_families(self, get_client, list_column_families):
        instance_method = get_client.return_value.instance
        instance_exists_method = instance_method.return_value.exists
        instance_exists_method.return_value = True
        client = mock.Mock(Client)
        get_client.return_value = client
        instance = google.cloud.bigtable.instance.Instance(instance_id=CBT_INSTANCE, client=client)
        self.bigtable_hook_default_project_id.get_column_families_for_table(
            instance=instance, table_id=CBT_TABLE
        )
        get_client.assert_not_called()
        list_column_families.assert_called_once_with()

    @mock.patch("google.cloud.bigtable.table.Table.get_cluster_states")
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_get_cluster_states(self, get_client, get_cluster_states):
        instance_method = get_client.return_value.instance
        instance_exists_method = instance_method.return_value.exists
        instance_exists_method.return_value = True
        client = mock.Mock(Client)
        instance = google.cloud.bigtable.instance.Instance(instance_id=CBT_INSTANCE, client=client)
        self.bigtable_hook_default_project_id.get_cluster_states_for_table(
            instance=instance, table_id=CBT_TABLE
        )
        get_client.assert_not_called()
        get_cluster_states.assert_called_once_with()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Example Airflow DAG for Google Gen AI Gemini Batch API and File API.
"""

from __future__ import annotations

import os
from datetime import datetime

try:
    from airflow.sdk import task
except ImportError:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from pathlib import Path

from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.gen_ai import (
    GenAIGeminiCancelBatchJobOperator,
    GenAIGeminiCreateBatchJobOperator,
    GenAIGeminiCreateEmbeddingsBatchJobOperator,
    GenAIGeminiDeleteBatchJobOperator,
    GenAIGeminiDeleteFileOperator,
    GenAIGeminiGetBatchJobOperator,
    GenAIGeminiGetFileOperator,
    GenAIGeminiListBatchJobsOperator,
    GenAIGeminiListFilesOperator,
    GenAIGeminiUploadFileOperator,
)
from airflow.providers.google.common.utils.get_secret import get_secret
from airflow.providers.standard.operators.bash import BashOperator

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT", "default")
REGION = "us-central1"
DAG_ID = "gen_ai_gemini_batch_api"

GEMINI_API_KEY = "api_key"

INLINED_REQUESTS_FOR_BATCH_JOB = [
    {"contents": [{"parts": [{"text": "Tell me a one-sentence joke."}], "role": "user"}]},
    {"contents": [{"parts": [{"text": "Why is the sky blue?"}], "role": "user"}]},
]

INLINED_REQUESTS_FOR_EMBEDDINGS_BATCH_JOB = {
    "contents": [{"parts": [{"text": "Why is the sky blue?"}], "role": "user"}]
}


GEMINI_XCOM_API_KEY = "{{ task_instance.xcom_pull('get_gemini_api_key') }}"

LOCAL_FILE_NAME = "gemini_batch_requests.jsonl"
LOCAL_EMBEDDINGS_FILE_NAME = "gemini_batch_embeddings_requests.jsonl"
UPLOAD_FILE_PATH = str(Path(__file__).parent / "resources" / LOCAL_FILE_NAME)
PATH_TO_SAVE_RESULTS = str(Path(__file__).parent / "resources")
UPLOAD_EMBEDDINGS_FILE_PATH = str(Path(__file__).parent / "resources" / LOCAL_EMBEDDINGS_FILE_NAME)

UPLOADED_FILE_NAME = (
    "{{ task_instance.xcom_pull(task_ids='upload_file_for_batch_job_task', key='file_name') }}"
)
UPLOADED_EMBEDDINGS_FILE_NAME = (
    "{{ task_instance.xcom_pull(task_ids='upload_file_for_embeddings_batch_job_task', key='file_name') }}"
)

BATCH_JOB_WITH_INLINED_REQUESTS_NAME = (
    "{{ task_instance.xcom_pull(task_ids='create_batch_job_using_inlined_requests_task', key='job_name') }}"
)
BATCH_JOB_WITH_FILE_NAME = (
    "{{ task_instance.xcom_pull(task_ids='create_batch_job_using_file_task', key='job_name') }}"
)
EMBEDDINGS_BATCH_JOB_WITH_INLINED_REQUESTS_NAME = "{{ task_instance.xcom_pull(task_ids='create_embeddings_job_using_inlined_requests_task', key='job_name') }}"
EMBEDDINGS_BATCH_JOB_WITH_FILE_NAME = (
    "{{ task_instance.xcom_pull(task_ids='create_embeddings_job_using_file_task', key='job_name') }}"
)


with DAG(
    dag_id=DAG_ID,
    description="Sample DAG with Gemini Batch API.",
    schedule="@once",
    start_date=datetime(2024, 1, 1),
    catchup=False,
    tags=["example", "gen_ai", "gemini_batch_api", "gemini_file_api"],
    render_template_as_native_obj=True,
) as dag:

    @task
    def get_gemini_api_key():
        return get_secret(GEMINI_API_KEY)

    get_gemini_api_key_task = get_gemini_api_key()

    # [START how_to_cloud_gen_ai_files_api_upload_file_task]
    upload_file = GenAIGeminiUploadFileOperator(
        task_id="upload_file_for_batch_job_task",
        project_id=PROJECT_ID,
        location=REGION,
        file_path=UPLOAD_FILE_PATH,
        gemini_api_key=GEMINI_XCOM_API_KEY,
    )
    # [END how_to_cloud_gen_ai_files_api_upload_file_task]

    upload_embeddings_file = GenAIGeminiUploadFileOperator(
        task_id="upload_file_for_embeddings_batch_job_task",
        project_id=PROJECT_ID,
        location=REGION,
        file_path=UPLOAD_EMBEDDINGS_FILE_PATH,
        gemini_api_key=GEMINI_XCOM_API_KEY,
    )

    # [START how_to_cloud_gen_ai_files_api_get_file_task]
    get_file = GenAIGeminiGetFileOperator(
        task_id="get_file_for_batch_job_task",
        project_id=PROJECT_ID,
        location=REGION,
        file_name=UPLOADED_FILE_NAME,
        gemini_api_key=GEMINI_XCOM_API_KEY,
    )
    # [END how_to_cloud_gen_ai_files_api_get_file_task]

    # [START how_to_cloud_gen_ai_files_api_list_files_task]
    list_files = GenAIGeminiListFilesOperator(
        task_id="list_files_files_api_task",
        project_id=PROJECT_ID,
        location=REGION,
        gemini_api_key=GEMINI_XCOM_API_KEY,
    )
    # [END how_to_cloud_gen_ai_files_api_list_files_task]

    # [START how_to_cloud_gen_ai_files_api_delete_file_task]
    delete_file = GenAIGeminiDeleteFileOperator(
        task_id="delete_file_for_batch_job_task",
        project_id=PROJECT_ID,
        location=REGION,
        file_name=UPLOADED_FILE_NAME,
        gemini_api_key=GEMINI_XCOM_API_KEY,
        trigger_rule=TriggerRule.ALL_DONE,
    )
    # [END how_to_cloud_gen_ai_files_api_delete_file_task]

    delete_embeddings_file = GenAIGeminiDeleteFileOperator(
        task_id="delete_file_for_embeddings_batch_job_task",
        project_id=PROJECT_ID,
        location=REGION,
        file_name=UPLOADED_EMBEDDINGS_FILE_NAME,
        gemini_api_key=GEMINI_XCOM_API_KEY,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    # [START how_to_cloud_gen_ai_batch_api_create_batch_job_with_inlined_requests_task]
    create_batch_job_using_inlined_requests = GenAIGeminiCreateBatchJobOperator(
        task_id="create_batch_job_using_inlined_requests_task",
        project_id=PROJECT_ID,
        location=REGION,
        model="gemini-3-pro-preview",
        gemini_api_key=GEMINI_XCOM_API_KEY,
        create_batch_job_config={
            "display_name": "inlined-requests-batch-job",
        },
        input_source=INLINED_REQUESTS_FOR_BATCH_JOB,
        wait_until_complete=True,
        retrieve_result=True,
    )
    # [END how_to_cloud_gen_ai_batch_api_create_batch_job_with_inlined_requests_task]

    # [START how_to_cloud_gen_ai_batch_api_create_batch_job_with_file_task]
    create_batch_job_using_file = GenAIGeminiCreateBatchJobOperator(
        task_id="create_batch_job_using_file_task",
        project_id=PROJECT_ID,
        location=REGION,
        model="gemini-3-pro-preview",
        gemini_api_key=GEMINI_XCOM_API_KEY,
        create_batch_job_config={
            "display_name": "file-upload-batch-job",
        },
        input_source=UPLOADED_FILE_NAME,
        wait_until_complete=True,
        retrieve_result=True,
        results_folder=PATH_TO_SAVE_RESULTS,
    )
    # [END how_to_cloud_gen_ai_batch_api_create_batch_job_with_file_task]

    # [START how_to_cloud_gen_ai_batch_api_create_embeddings_with_inlined_requests_task]
    create_embeddings_job_using_inlined_requests = GenAIGeminiCreateEmbeddingsBatchJobOperator(
        task_id="create_embeddings_job_using_inlined_requests_task",
        project_id=PROJECT_ID,
        location=REGION,
        model="gemini-embedding-001",
        wait_until_complete=False,
        gemini_api_key=GEMINI_XCOM_API_KEY,
        create_embeddings_config={
            "display_name": "inlined-requests-embeddings-job",
        },
        input_source=INLINED_REQUESTS_FOR_EMBEDDINGS_BATCH_JOB,
    )
    # [END how_to_cloud_gen_ai_batch_api_create_embeddings_with_inlined_requests_task]

    # [START how_to_cloud_gen_ai_batch_api_create_embeddings_with_file_task]
    create_embeddings_job_using_file = GenAIGeminiCreateEmbeddingsBatchJobOperator(
        task_id="create_embeddings_job_using_file_task",
        project_id=PROJECT_ID,
        location=REGION,
        model="gemini-embedding-001",
        wait_until_complete=False,
        gemini_api_key=GEMINI_XCOM_API_KEY,
        create_embeddings_config={
            "display_name": "file-upload-embeddings-job",
        },
        input_source=UPLOADED_EMBEDDINGS_FILE_NAME,
    )
    # [END how_to_cloud_gen_ai_batch_api_create_embeddings_with_file_task]

    # [START how_to_cloud_gen_ai_batch_api_get_batch_job_task]
    get_batch_job = GenAIGeminiGetBatchJobOperator(
        task_id="get_batch_job_task",
        project_id=PROJECT_ID,
        location=REGION,
        gemini_api_key=GEMINI_XCOM_API_KEY,
        job_name=BATCH_JOB_WITH_INLINED_REQUESTS_NAME,
    )
    # [END how_to_cloud_gen_ai_batch_api_get_batch_job_task]

    # [START how_to_cloud_gen_ai_batch_api_list_batch_jobs_task]
    list_batch_jobs = GenAIGeminiListBatchJobsOperator(
        task_id="list_batch_jobs_task",
        project_id=PROJECT_ID,
        location=REGION,
        gemini_api_key=GEMINI_XCOM_API_KEY,
    )
    # [END how_to_cloud_gen_ai_batch_api_list_batch_jobs_task]

    # [START how_to_cloud_gen_ai_batch_api_cancel_batch_job_task]
    cancel_batch_job = GenAIGeminiCancelBatchJobOperator(
        task_id="cancel_batch_job_task",
        project_id=PROJECT_ID,
        location=REGION,
        gemini_api_key=GEMINI_XCOM_API_KEY,
        job_name=BATCH_JOB_WITH_FILE_NAME,
    )
    # [END how_to_cloud_gen_ai_batch_api_cancel_batch_job_task]

    # [START how_to_cloud_gen_ai_batch_api_delete_batch_job_task]
    delete_batch_job_1 = GenAIGeminiDeleteBatchJobOperator(
        task_id="delete_batch_job_1_task",
        project_id=PROJECT_ID,
        location=REGION,
        gemini_api_key=GEMINI_XCOM_API_KEY,
        job_name=BATCH_JOB_WITH_INLINED_REQUESTS_NAME,
        trigger_rule=TriggerRule.ALL_DONE,
    )
    # [END how_to_cloud_gen_ai_batch_api_delete_batch_job_task]

    delete_batch_job_2 = GenAIGeminiDeleteBatchJobOperator(
        task_id="delete_batch_job_2_task",
        project_id=PROJECT_ID,
        location=REGION,
        gemini_api_key=GEMINI_XCOM_API_KEY,
        job_name=BATCH_JOB_WITH_FILE_NAME,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    delete_embeddings_batch_job_1 = GenAIGeminiDeleteBatchJobOperator(
        task_id="delete_embeddings_batch_job_1_task",
        project_id=PROJECT_ID,
        location=REGION,
        gemini_api_key=GEMINI_XCOM_API_KEY,
        job_name=EMBEDDINGS_BATCH_JOB_WITH_INLINED_REQUESTS_NAME,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    delete_embeddings_batch_job_2 = GenAIGeminiDeleteBatchJobOperator(
        task_id="delete_embeddings_batch_job_2_task",
        project_id=PROJECT_ID,
        location=REGION,
        gemini_api_key=GEMINI_XCOM_API_KEY,
        job_name=EMBEDDINGS_BATCH_JOB_WITH_FILE_NAME,
        trigger_rule=TriggerRule.ALL_DONE,
    )
    delete_result_file = BashOperator(
        task_id="delete_result_file_task",
        bash_command="rm "
        "{{ task_instance.xcom_pull(task_ids='create_batch_job_using_file_task', key='job_results') }}",
        trigger_rule=TriggerRule.ALL_DONE,
    )

    (
        get_gemini_api_key_task
        >> [upload_file, upload_embeddings_file]
        >> get_file
        >> list_files
        >> [
            create_batch_job_using_inlined_requests,
            create_batch_job_using_file,
            create_embeddings_job_using_file,
            create_embeddings_job_using_inlined_requests,
        ]
        >> get_batch_job
        >> list_batch_jobs
        >> cancel_batch_job
        >> delete_batch_job_1
        >> delete_batch_job_2
        >> delete_embeddings_batch_job_1
        >> delete_embeddings_batch_job_2
        >> [delete_file, delete_embeddings_file, delete_result_file]
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains Google Vertex AI Ray operators."""

from __future__ import annotations

from collections.abc import Sequence
from functools import cached_property
from typing import TYPE_CHECKING, Literal

from google.api_core.exceptions import NotFound
from google.cloud.aiplatform.vertex_ray.util import resources

from airflow.providers.common.compat.sdk import AirflowException
from airflow.providers.google.cloud.hooks.vertex_ai.ray import RayHook
from airflow.providers.google.cloud.links.vertex_ai import (
    VertexAIRayClusterLink,
    VertexAIRayClusterListLink,
)
from airflow.providers.google.cloud.operators.cloud_base import GoogleCloudBaseOperator

if TYPE_CHECKING:
    from airflow.providers.common.compat.sdk import Context


class RayBaseOperator(GoogleCloudBaseOperator):
    """
    Base class for Ray operators.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param location: Required. The ID of the Google Cloud region that the service belongs to.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "location",
        "gcp_conn_id",
        "project_id",
        "impersonation_chain",
    )

    def __init__(
        self,
        project_id: str,
        location: str,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.location = location
        self.project_id = project_id
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    @cached_property
    def hook(self) -> RayHook:
        return RayHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )


class CreateRayClusterOperator(RayBaseOperator):
    """
    Create a Ray cluster on the Vertex AI.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param location: Required. The ID of the Google Cloud region that the service belongs to.
    :param head_node_type: The head node resource. Resources.node_count must be 1. If not set, default
        value of Resources() class will be used.
    :param python_version: Required. Python version for the ray cluster.
    :param ray_version: Required. Ray version for the ray cluster.
        Currently only 3 version are available: 2.9.3, 2.33, 2.42. For more information please refer to
        https://github.com/googleapis/python-aiplatform/blob/main/setup.py#L101
    :param network: Virtual private cloud (VPC) network. For Ray Client, VPC peering is required to
        connect to the Ray Cluster managed in the Vertex API service. For Ray Job API, VPC network is not
        required because Ray Cluster connection can be accessed through dashboard address.
    :param service_account: Service account to be used for running Ray programs on the cluster.
    :param cluster_name: This value may be up to 63 characters, and valid characters are `[a-z0-9_-]`.
        The first character cannot be a number or hyphen.
    :param worker_node_types: The list of Resources of the worker nodes. The same Resources object should
        not appear multiple times in the list.
    :param custom_images: The NodeImages which specifies head node and worker nodes images. All the
        workers will share the same image. If each Resource has a specific custom image, use
        `Resources.custom_image` for head/worker_node_type(s). Note that configuring
        `Resources.custom_image` will override `custom_images` here. Allowlist only.
    :param enable_metrics_collection: Enable Ray metrics collection for visualization.
    :param enable_logging: Enable exporting Ray logs to Cloud Logging.
    :param psc_interface_config: PSC-I config.
    :param reserved_ip_ranges: A list of names for the reserved IP ranges under the VPC network that can
        be used for this cluster. If set, we will deploy the cluster within the provided IP ranges.
        Otherwise, the cluster is deployed to any IP ranges under the provided VPC network.
        Example: ["vertex-ai-ip-range"].
    :param labels: The labels with user-defined metadata to organize Ray cluster.
        Label keys and values can be no longer than 64 characters (Unicode codepoints), can only contain
        lowercase letters, numeric characters, underscores and dashes. International characters are allowed.
        See https://goo.gl/xmQnxf for more information and examples of labels.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = tuple(
        {"head_node_type", "worker_node_types"} | set(RayBaseOperator.template_fields)
    )
    operator_extra_links = (VertexAIRayClusterLink(),)

    def __init__(
        self,
        python_version: str,
        ray_version: Literal["2.9.3", "2.33", "2.42"],
        head_node_type: resources.Resources = resources.Resources(),
        network: str | None = None,
        service_account: str | None = None,
        cluster_name: str | None = None,
        worker_node_types: list[resources.Resources] | None = None,
        custom_images: resources.NodeImages | None = None,
        enable_metrics_collection: bool = True,
        enable_logging: bool = True,
        psc_interface_config: resources.PscIConfig | None = None,
        reserved_ip_ranges: list[str] | None = None,
        labels: dict[str, str] | None = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.head_node_type = head_node_type
        self.python_version = python_version
        self.ray_version = ray_version
        self.network = network
        self.service_account = service_account
        self.cluster_name = cluster_name
        self.worker_node_types = worker_node_types
        self.custom_images = custom_images
        self.enable_metrics_collection = enable_metrics_collection
        self.enable_logging = enable_logging
        self.psc_interface_config = psc_interface_config
        self.reserved_ip_ranges = reserved_ip_ranges
        self.labels = labels

    def execute(self, context: Context):
        self.log.info("Creating a Ray cluster.")
        try:
            cluster_path = self.hook.create_ray_cluster(
                project_id=self.project_id,
                location=self.location,
                head_node_type=self.head_node_type,
                python_version=self.python_version,
                ray_version=self.ray_version,
                network=self.network,
                service_account=self.service_account,
                cluster_name=self.cluster_name,
                worker_node_types=self.worker_node_types,
                custom_images=self.custom_images,
                enable_metrics_collection=self.enable_metrics_collection,
                enable_logging=self.enable_logging,
                psc_interface_config=self.psc_interface_config,
                reserved_ip_ranges=self.reserved_ip_ranges,
                labels=self.labels,
            )
            cluster_id = self.hook.extract_cluster_id(cluster_path)
            context["ti"].xcom_push(
                key="cluster_id",
                value=cluster_id,
            )
            VertexAIRayClusterLink.persist(
                context=context, location=self.location, cluster_id=cluster_id, project_id=self.project_id
            )
            self.log.info("Ray cluster was created.")
        except Exception as error:
            raise AirflowException(error)
        return cluster_path


class ListRayClustersOperator(RayBaseOperator):
    """
    List Ray clusters under the currently authenticated project.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param location: Required. The ID of the Google Cloud region that the service belongs to.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    operator_extra_links = (VertexAIRayClusterListLink(),)

    def execute(self, context: Context):
        VertexAIRayClusterListLink.persist(context=context, project_id=self.project_id)
        self.log.info("Listing Clusters from location %s.", self.location)
        try:
            ray_cluster_list = self.hook.list_ray_clusters(
                project_id=self.project_id,
                location=self.location,
            )
            ray_cluster_dict_list = [
                self.hook.serialize_cluster_obj(ray_cluster) for ray_cluster in ray_cluster_list
            ]
        except Exception as error:
            raise AirflowException(error)
        return ray_cluster_dict_list


class GetRayClusterOperator(RayBaseOperator):
    """
    Get Ray cluster.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param location: Required. The ID of the Google Cloud region that the service belongs to.
    :param cluster_id: Cluster resource ID.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = tuple({"cluster_id"} | set(RayBaseOperator.template_fields))
    operator_extra_links = (VertexAIRayClusterLink(),)

    def __init__(
        self,
        cluster_id: str,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.cluster_id = cluster_id

    def execute(self, context: Context):
        VertexAIRayClusterLink.persist(
            context=context,
            location=self.location,
            cluster_id=self.cluster_id,
            project_id=self.project_id,
        )
        self.log.info("Getting Cluster: %s", self.cluster_id)
        try:
            ray_cluster = self.hook.get_ray_cluster(
                project_id=self.project_id,
                location=self.location,
                cluster_id=self.cluster_id,
            )
            self.log.info("Cluster data has been retrieved.")
            ray_cluster_dict = self.hook.serialize_cluster_obj(ray_cluster)
            return ray_cluster_dict
        except NotFound as not_found_err:
            self.log.info("The Cluster %s does not exist.", self.cluster_id)
            raise AirflowException(not_found_err)


class UpdateRayClusterOperator(RayBaseOperator):
    """
    Update Ray cluster (currently support resizing node counts for worker nodes).

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param location: Required. The ID of the Google Cloud region that the service belongs to.
    :param cluster_id: Cluster resource ID.
    :param worker_node_types: The list of Resources of the resized worker nodes. The same Resources
        object should not appear multiple times in the list.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = tuple(
        {"cluster_id", "worker_node_types"} | set(RayBaseOperator.template_fields)
    )
    operator_extra_links = (VertexAIRayClusterLink(),)

    def __init__(
        self,
        cluster_id: str,
        worker_node_types: list[resources.Resources],
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.cluster_id = cluster_id
        self.worker_node_types = worker_node_types

    def execute(self, context: Context):
        VertexAIRayClusterLink.persist(
            context=context,
            location=self.location,
            cluster_id=self.cluster_id,
            project_id=self.project_id,
        )
        self.log.info("Updating a Ray cluster.")
        try:
            cluster_path = self.hook.update_ray_cluster(
                project_id=self.project_id,
                location=self.location,
                cluster_id=self.cluster_id,
                worker_node_types=self.worker_node_types,
            )
            self.log.info("Ray cluster %s was updated.", self.cluster_id)
            return cluster_path
        except NotFound as not_found_err:
            self.log.info("The Cluster %s does not exist.", self.cluster_id)
            raise AirflowException(not_found_err)
        except Exception as error:
            raise AirflowException(error)


class DeleteRayClusterOperator(RayBaseOperator):
    """
    Delete Ray cluster.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param location: Required. The ID of the Google Cloud region that the service belongs to.
    :param cluster_id: Cluster resource ID.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = tuple({"cluster_id"} | set(RayBaseOperator.template_fields))

    def __init__(
        self,
        cluster_id: str,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.cluster_id = cluster_id

    def execute(self, context: Context):
        try:
            self.log.info("Deleting Ray cluster: %s", self.cluster_id)
            self.hook.delete_ray_cluster(
                project_id=self.project_id,
                location=self.location,
                cluster_id=self.cluster_id,
            )
            self.log.info("Ray cluster was deleted.")
        except NotFound as not_found_err:
            self.log.info("The Ray cluster ID %s does not exist.", self.cluster_id)
            raise AirflowException(not_found_err)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import MagicMock

import pytest

from airflow.models import Connection
from airflow.providers.common.compat.sdk import AirflowException
from airflow.providers.common.sql.hooks.sql import DbApiHook
from airflow.providers.google.cloud.operators.spanner import (
    SpannerDeleteDatabaseInstanceOperator,
    SpannerDeleteInstanceOperator,
    SpannerDeployDatabaseInstanceOperator,
    SpannerDeployInstanceOperator,
    SpannerQueryDatabaseInstanceOperator,
    SpannerUpdateDatabaseInstanceOperator,
)
from airflow.providers.openlineage.sqlparser import DatabaseInfo

PROJECT_ID = "project-id"
INSTANCE_ID = "instance-id"
DB_ID = "db1"
CONFIG_NAME = "projects/project-id/instanceConfigs/eur3"
NODE_COUNT = "1"
DISPLAY_NAME = "Test Instance"
INSERT_QUERY = "INSERT my_table1 (id, name) VALUES (1, 'One')"
INSERT_QUERY_2 = "INSERT my_table2 (id, name) VALUES (1, 'One')"
CREATE_QUERY = "CREATE TABLE my_table1 (id INT64, name STRING(100))"
CREATE_QUERY_2 = "CREATE TABLE my_table2 (id INT64, name STRING(100))"
DDL_STATEMENTS = [CREATE_QUERY, CREATE_QUERY_2]
TASK_ID = "task-id"

SCHEMA_ROWS = {
    "public.orders": [
        ("public", "orders", "id", 1, "INT64"),
        ("public", "orders", "amount", 2, "FLOAT64"),
    ],
    "public.staging": [
        ("public", "staging", "id", 1, "INT64"),
        ("public", "staging", "amount", 2, "FLOAT64"),
    ],
    "public.customers": [
        ("public", "customers", "id", 1, "INT64"),
        ("public", "customers", "name", 2, "STRING(100)"),
        ("public", "customers", "customer_id", 3, "INT64"),
    ],
    "public.logs": [
        ("public", "logs", "id", 1, "INT64"),
        ("public", "logs", "message", 2, "STRING(100)"),
    ],
    "public.t1": [("public", "t1", "col1", 1, "STRING(100)")],
    "public.t2": [("public", "t2", "col1", 1, "STRING(100)")],
    "public.t3": [("public", "t3", "id", 1, "INT64")],
    # example of explicit non-default schema
    "myschema.orders": [("myschema", "orders", "id", 1, "INT64")],
}


class TestCloudSpanner:
    @mock.patch("airflow.providers.google.cloud.operators.spanner.SpannerHook")
    def test_instance_create(self, mock_hook):
        mock_hook.return_value.get_instance.return_value = None
        op = SpannerDeployInstanceOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            configuration_name=CONFIG_NAME,
            node_count=int(NODE_COUNT),
            display_name=DISPLAY_NAME,
            task_id="id",
        )
        context = mock.MagicMock()
        result = op.execute(context=context)
        mock_hook.assert_called_once_with(
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.create_instance.assert_called_once_with(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            configuration_name=CONFIG_NAME,
            node_count=int(NODE_COUNT),
            display_name=DISPLAY_NAME,
        )
        mock_hook.return_value.update_instance.assert_not_called()
        assert result is None

    @mock.patch("airflow.providers.google.cloud.operators.spanner.SpannerHook")
    def test_instance_create_missing_project_id(self, mock_hook):
        mock_hook.return_value.get_instance.return_value = None
        op = SpannerDeployInstanceOperator(
            instance_id=INSTANCE_ID,
            configuration_name=CONFIG_NAME,
            node_count=int(NODE_COUNT),
            display_name=DISPLAY_NAME,
            task_id="id",
        )
        context = mock.MagicMock()
        result = op.execute(context=context)
        mock_hook.assert_called_once_with(
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.create_instance.assert_called_once_with(
            project_id=None,
            instance_id=INSTANCE_ID,
            configuration_name=CONFIG_NAME,
            node_count=int(NODE_COUNT),
            display_name=DISPLAY_NAME,
        )
        mock_hook.return_value.update_instance.assert_not_called()
        assert result is None

    @mock.patch("airflow.providers.google.cloud.operators.spanner.SpannerHook")
    def test_instance_update(self, mock_hook):
        mock_hook.return_value.get_instance.return_value = {"name": INSTANCE_ID}
        op = SpannerDeployInstanceOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            configuration_name=CONFIG_NAME,
            node_count=int(NODE_COUNT),
            display_name=DISPLAY_NAME,
            task_id="id",
        )
        context = mock.MagicMock()
        result = op.execute(context=context)
        mock_hook.assert_called_once_with(
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.update_instance.assert_called_once_with(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            configuration_name=CONFIG_NAME,
            node_count=int(NODE_COUNT),
            display_name=DISPLAY_NAME,
        )
        mock_hook.return_value.create_instance.assert_not_called()
        assert result is None

    @mock.patch("airflow.providers.google.cloud.operators.spanner.SpannerHook")
    def test_instance_update_missing_project_id(self, mock_hook):
        mock_hook.return_value.get_instance.return_value = {"name": INSTANCE_ID}
        op = SpannerDeployInstanceOperator(
            instance_id=INSTANCE_ID,
            configuration_name=CONFIG_NAME,
            node_count=int(NODE_COUNT),
            display_name=DISPLAY_NAME,
            task_id="id",
        )
        context = mock.MagicMock()
        result = op.execute(context=context)
        mock_hook.assert_called_once_with(
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.update_instance.assert_called_once_with(
            project_id=None,
            instance_id=INSTANCE_ID,
            configuration_name=CONFIG_NAME,
            node_count=int(NODE_COUNT),
            display_name=DISPLAY_NAME,
        )
        mock_hook.return_value.create_instance.assert_not_called()
        assert result is None

    @mock.patch("airflow.providers.google.cloud.operators.spanner.SpannerHook")
    def test_instance_create_aborts_and_succeeds_if_instance_exists(self, mock_hook):
        mock_hook.return_value.get_instance.return_value = {"name": INSTANCE_ID}
        op = SpannerDeployInstanceOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            configuration_name=CONFIG_NAME,
            node_count=int(NODE_COUNT),
            display_name=DISPLAY_NAME,
            task_id="id",
        )
        context = mock.MagicMock()
        result = op.execute(context=context)
        mock_hook.assert_called_once_with(
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.create_instance.assert_not_called()
        assert result is None

    @pytest.mark.parametrize(
        ("project_id", "instance_id", "exp_msg"),
        [
            ("", INSTANCE_ID, "project_id"),
            (PROJECT_ID, "", "instance_id"),
        ],
    )
    @mock.patch("airflow.providers.google.cloud.operators.spanner.SpannerHook")
    def test_instance_create_ex_if_param_missing(self, mock_hook, project_id, instance_id, exp_msg):
        with pytest.raises(AirflowException) as ctx:
            SpannerDeployInstanceOperator(
                project_id=project_id,
                instance_id=instance_id,
                configuration_name=CONFIG_NAME,
                node_count=int(NODE_COUNT),
                display_name=DISPLAY_NAME,
                task_id="id",
            )
        err = ctx.value
        assert f"The required parameter '{exp_msg}' is empty" in str(err)
        mock_hook.assert_not_called()

    @mock.patch("airflow.providers.google.cloud.operators.spanner.SpannerHook")
    def test_instance_delete(self, mock_hook):
        mock_hook.return_value.get_instance.return_value = {"name": INSTANCE_ID}
        op = SpannerDeleteInstanceOperator(project_id=PROJECT_ID, instance_id=INSTANCE_ID, task_id="id")
        result = op.execute(None)
        mock_hook.assert_called_once_with(
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.delete_instance.assert_called_once_with(
            project_id=PROJECT_ID, instance_id=INSTANCE_ID
        )
        assert result

    @mock.patch("airflow.providers.google.cloud.operators.spanner.SpannerHook")
    def test_instance_delete_missing_project_id(self, mock_hook):
        mock_hook.return_value.get_instance.return_value = {"name": INSTANCE_ID}
        op = SpannerDeleteInstanceOperator(instance_id=INSTANCE_ID, task_id="id")
        result = op.execute(None)
        mock_hook.assert_called_once_with(
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.delete_instance.assert_called_once_with(
            project_id=None, instance_id=INSTANCE_ID
        )
        assert result

    @mock.patch("airflow.providers.google.cloud.operators.spanner.SpannerHook")
    def test_instance_delete_aborts_and_succeeds_if_instance_does_not_exist(self, mock_hook):
        mock_hook.return_value.get_instance.return_value = None
        op = SpannerDeleteInstanceOperator(project_id=PROJECT_ID, instance_id=INSTANCE_ID, task_id="id")
        result = op.execute(None)
        mock_hook.assert_called_once_with(
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.delete_instance.assert_not_called()
        assert result

    @pytest.mark.parametrize(
        ("project_id", "instance_id", "exp_msg"),
        [
            ("", INSTANCE_ID, "project_id"),
            (PROJECT_ID, "", "instance_id"),
        ],
    )
    @mock.patch("airflow.providers.google.cloud.operators.spanner.SpannerHook")
    def test_instance_delete_ex_if_param_missing(self, mock_hook, project_id, instance_id, exp_msg):
        with pytest.raises(AirflowException) as ctx:
            SpannerDeleteInstanceOperator(project_id=project_id, instance_id=instance_id, task_id="id")
        err = ctx.value
        assert f"The required parameter '{exp_msg}' is empty" in str(err)
        mock_hook.assert_not_called()

    @mock.patch("airflow.providers.google.cloud.operators.spanner.SpannerHook")
    def test_instance_query(self, mock_hook):
        mock_hook.return_value.execute_dml.return_value = [3]
        op = SpannerQueryDatabaseInstanceOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            database_id=DB_ID,
            query=INSERT_QUERY,
            task_id="id",
        )
        result = op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.execute_dml.assert_called_once_with(
            project_id=PROJECT_ID, instance_id=INSTANCE_ID, database_id=DB_ID, queries=[INSERT_QUERY]
        )
        assert result == [3]

    @mock.patch("airflow.providers.google.cloud.operators.spanner.SpannerHook")
    def test_instance_query_missing_project_id(self, mock_hook):
        mock_hook.return_value.execute_dml.return_value = [3]
        op = SpannerQueryDatabaseInstanceOperator(
            instance_id=INSTANCE_ID, database_id=DB_ID, query=INSERT_QUERY, task_id="id"
        )
        context = mock.MagicMock()
        result = op.execute(context=context)
        mock_hook.assert_called_once_with(
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.execute_dml.assert_called_once_with(
            project_id=None, instance_id=INSTANCE_ID, database_id=DB_ID, queries=[INSERT_QUERY]
        )
        assert result == [3]

    @pytest.mark.parametrize(
        ("project_id", "instance_id", "database_id", "query", "exp_msg"),
        [
            ("", INSTANCE_ID, DB_ID, INSERT_QUERY, "project_id"),
            (PROJECT_ID, "", DB_ID, INSERT_QUERY, "instance_id"),
            (PROJECT_ID, INSTANCE_ID, "", INSERT_QUERY, "database_id"),
            (PROJECT_ID, INSTANCE_ID, DB_ID, "", "query"),
        ],
    )
    @mock.patch("airflow.providers.google.cloud.operators.spanner.SpannerHook")
    def test_instance_query_ex_if_param_missing(
        self, mock_hook, project_id, instance_id, database_id, query, exp_msg
    ):
        with pytest.raises(AirflowException) as ctx:
            SpannerQueryDatabaseInstanceOperator(
                project_id=project_id,
                instance_id=instance_id,
                database_id=database_id,
                query=query,
                task_id="id",
            )
        err = ctx.value
        assert f"The required parameter '{exp_msg}' is empty" in str(err)
        mock_hook.assert_not_called()

    @mock.patch("airflow.providers.google.cloud.operators.spanner.SpannerHook")
    def test_instance_query_dml(self, mock_hook):
        mock_hook.return_value.execute_dml.return_value = None
        op = SpannerQueryDatabaseInstanceOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            database_id=DB_ID,
            query=INSERT_QUERY,
            task_id="id",
        )
        context = mock.MagicMock()
        op.execute(context=context)
        mock_hook.assert_called_once_with(
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.execute_dml.assert_called_once_with(
            project_id=PROJECT_ID, instance_id=INSTANCE_ID, database_id=DB_ID, queries=[INSERT_QUERY]
        )

    @mock.patch("airflow.providers.google.cloud.operators.spanner.SpannerHook")
    def test_instance_query_dml_list(self, mock_hook):
        mock_hook.return_value.execute_dml.return_value = None
        op = SpannerQueryDatabaseInstanceOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            database_id=DB_ID,
            query=[INSERT_QUERY, INSERT_QUERY_2],
            task_id="id",
        )
        context = mock.MagicMock()
        op.execute(context=context)
        mock_hook.assert_called_once_with(
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.execute_dml.assert_called_once_with(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            database_id=DB_ID,
            queries=[INSERT_QUERY, INSERT_QUERY_2],
        )

    @pytest.mark.parametrize(
        ("sql", "expected_inputs", "expected_outputs", "expected_lineage"),
        [
            ("SELECT id, amount FROM public.orders", ["db1.public.orders"], [], {}),
            (
                "INSERT INTO public.orders (id, amount) SELECT id, amount FROM public.staging",
                ["db1.public.staging", "db1.public.orders"],
                [],
                {},
            ),
            ("DELETE FROM public.logs WHERE id=1", [], ["db1.public.logs"], {}),
            (
                "SELECT o.id, c.name FROM public.orders o JOIN public.customers c ON o.customer_id = c.id",
                ["db1.public.orders", "db1.public.customers"],
                [],
                {},
            ),
            (
                "UPDATE public.customers SET name='x' WHERE id IN (SELECT id FROM public.staging)",
                ["db1.public.customers", "db1.public.staging"],
                [],
                {},
            ),
            (
                ["INSERT INTO public.t1 SELECT * FROM public.t2;", "DELETE FROM public.t3 WHERE id=1;"],
                ["db1.public.t1", "db1.public.t2", "db1.public.t3"],
                [],
                {},
            ),
            ("SELECT id, amount FROM myschema.orders", ["db1.myschema.orders"], [], {}),
        ],
    )
    def test_spannerquerydatabaseinstanceoperator_get_openlineage_facets(
        self, sql, expected_inputs, expected_outputs, expected_lineage
    ):
        # Arrange
        class SpannerHookForTests(DbApiHook):
            conn_name_attr = "gcp_conn_id"
            get_conn = MagicMock(name="conn")
            get_connection = MagicMock()
            database = DB_ID

            def get_openlineage_database_info(self, connection):
                return DatabaseInfo(
                    scheme="spanner",
                    authority=f"{PROJECT_ID}/{INSTANCE_ID}",
                    database=DB_ID,
                    information_schema_columns=[
                        "table_schema",
                        "table_name",
                        "column_name",
                        "ordinal_position",
                        "spanner_type",
                    ],
                    information_schema_table_name="information_schema.columns",
                    use_flat_cross_db_query=False,
                    is_information_schema_cross_db=False,
                    is_uppercase_names=False,
                )

        dbapi_hook = SpannerHookForTests()

        class SpannerOperatorForTest(SpannerQueryDatabaseInstanceOperator):
            @property
            def hook(self):
                return dbapi_hook

        op = SpannerOperatorForTest(
            task_id=TASK_ID,
            instance_id=INSTANCE_ID,
            database_id=DB_ID,
            gcp_conn_id="spanner_conn",
            query=sql,
        )

        dbapi_hook.get_connection.return_value = Connection(
            conn_id="spanner_conn", conn_type="spanner", host="spanner-host"
        )

        combined_rows = []
        for ds in expected_inputs + expected_outputs:
            tbl = ds.split(".", 1)[1]
            combined_rows.extend(SCHEMA_ROWS.get(tbl, []))

        dbapi_hook.get_conn.return_value.cursor.return_value.fetchall.side_effect = [combined_rows, []]

        # Act
        lineage = op.get_openlineage_facets_on_complete(task_instance=None)
        assert lineage is not None

        # Assert inputs
        input_names = {ds.name for ds in lineage.inputs}
        assert input_names == set(expected_inputs)
        for ds in lineage.inputs:
            assert ds.namespace == f"spanner://{PROJECT_ID}/{INSTANCE_ID}"

        # Assert outputs
        output_names = {ds.name for ds in lineage.outputs}
        assert output_names == set(expected_outputs)
        for ds in lineage.outputs:
            assert ds.namespace == f"spanner://{PROJECT_ID}/{INSTANCE_ID}"

        # Assert SQLJobFacet
        sql_job = lineage.job_facets["sql"]
        if isinstance(sql, list):
            for q in sql:
                assert q.replace(";", "").strip() in sql_job.query.replace(";", "")
        else:
            assert sql_job.query == sql

        # Assert column lineage
        found_lineage = {
            getattr(field, "field", None) or getattr(field, "name", None): [
                f"{inp.dataset.name}.{getattr(inp, 'field', getattr(inp, 'name', None))}"
                for inp in getattr(field, "inputFields", [])
            ]
            for ds in lineage.outputs + lineage.inputs
            for cl_facet in [ds.facets.get("columnLineage")]
            if cl_facet
            for field in cl_facet.fields
        }

        for col, sources in expected_lineage.items():
            assert col in found_lineage
            for src in sources:
                assert any(src in s for s in found_lineage[col])

    @mock.patch("airflow.providers.google.cloud.operators.spanner.SpannerHook")
    def test_database_create(self, mock_hook):
        mock_hook.return_value.get_database.return_value = None
        op = SpannerDeployDatabaseInstanceOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            database_id=DB_ID,
            ddl_statements=DDL_STATEMENTS,
            task_id="id",
        )
        context = mock.MagicMock()
        result = op.execute(context=context)
        mock_hook.assert_called_once_with(
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.create_database.assert_called_once_with(
            project_id=PROJECT_ID, instance_id=INSTANCE_ID, database_id=DB_ID, ddl_statements=DDL_STATEMENTS
        )
        mock_hook.return_value.update_database.assert_not_called()
        assert result

    @mock.patch("airflow.providers.google.cloud.operators.spanner.SpannerHook")
    def test_database_create_missing_project_id(self, mock_hook):
        mock_hook.return_value.get_database.return_value = None
        op = SpannerDeployDatabaseInstanceOperator(
            instance_id=INSTANCE_ID, database_id=DB_ID, ddl_statements=DDL_STATEMENTS, task_id="id"
        )
        context = mock.MagicMock()
        result = op.execute(context=context)
        mock_hook.assert_called_once_with(
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.create_database.assert_called_once_with(
            project_id=None, instance_id=INSTANCE_ID, database_id=DB_ID, ddl_statements=DDL_STATEMENTS
        )
        mock_hook.return_value.update_database.assert_not_called()
        assert result

    @mock.patch("airflow.providers.google.cloud.operators.spanner.SpannerHook")
    def test_database_create_with_pre_existing_db(self, mock_hook):
        mock_hook.return_value.get_database.return_value = {"name": DB_ID}
        op = SpannerDeployDatabaseInstanceOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            database_id=DB_ID,
            ddl_statements=DDL_STATEMENTS,
            task_id="id",
        )
        context = mock.MagicMock()
        result = op.execute(context=context)
        mock_hook.assert_called_once_with(
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.create_database.assert_not_called()
        mock_hook.return_value.update_database.assert_not_called()
        assert result

    @pytest.mark.parametrize(
        ("project_id", "instance_id", "database_id", "ddl_statements", "exp_msg"),
        [
            ("", INSTANCE_ID, DB_ID, DDL_STATEMENTS, "project_id"),
            (PROJECT_ID, "", DB_ID, DDL_STATEMENTS, "instance_id"),
            (PROJECT_ID, INSTANCE_ID, "", DDL_STATEMENTS, "database_id"),
        ],
    )
    @mock.patch("airflow.providers.google.cloud.operators.spanner.SpannerHook")
    def test_database_create_ex_if_param_missing(
        self, mock_hook, project_id, instance_id, database_id, ddl_statements, exp_msg
    ):
        with pytest.raises(AirflowException) as ctx:
            SpannerDeployDatabaseInstanceOperator(
                project_id=project_id,
                instance_id=instance_id,
                database_id=database_id,
                ddl_statements=ddl_statements,
                task_id="id",
            )
        err = ctx.value
        assert f"The required parameter '{exp_msg}' is empty" in str(err)
        mock_hook.assert_not_called()

    @mock.patch("airflow.providers.google.cloud.operators.spanner.SpannerHook")
    def test_database_update(self, mock_hook):
        mock_hook.return_value.get_database.return_value = {"name": DB_ID}
        op = SpannerUpdateDatabaseInstanceOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            database_id=DB_ID,
            ddl_statements=DDL_STATEMENTS,
            task_id="id",
        )
        context = mock.MagicMock()
        result = op.execute(context=context)
        mock_hook.assert_called_once_with(
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.update_database.assert_called_once_with(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            database_id=DB_ID,
            ddl_statements=DDL_STATEMENTS,
            operation_id=None,
        )
        assert result

    @mock.patch("airflow.providers.google.cloud.operators.spanner.SpannerHook")
    def test_database_update_missing_project_id(self, mock_hook):
        mock_hook.return_value.get_database.return_value = {"name": DB_ID}
        op = SpannerUpdateDatabaseInstanceOperator(
            instance_id=INSTANCE_ID, database_id=DB_ID, ddl_statements=DDL_STATEMENTS, task_id="id"
        )
        context = mock.MagicMock()
        result = op.execute(context=context)
        mock_hook.assert_called_once_with(
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.update_database.assert_called_once_with(
            project_id=None,
            instance_id=INSTANCE_ID,
            database_id=DB_ID,
            ddl_statements=DDL_STATEMENTS,
            operation_id=None,
        )
        assert result

    @pytest.mark.parametrize(
        ("project_id", "instance_id", "database_id", "ddl_statements", "exp_msg"),
        [
            ("", INSTANCE_ID, DB_ID, DDL_STATEMENTS, "project_id"),
            (PROJECT_ID, "", DB_ID, DDL_STATEMENTS, "instance_id"),
            (PROJECT_ID, INSTANCE_ID, "", DDL_STATEMENTS, "database_id"),
        ],
    )
    @mock.patch("airflow.providers.google.cloud.operators.spanner.SpannerHook")
    def test_database_update_ex_if_param_missing(
        self, mock_hook, project_id, instance_id, database_id, ddl_statements, exp_msg
    ):
        with pytest.raises(AirflowException) as ctx:
            SpannerUpdateDatabaseInstanceOperator(
                project_id=project_id,
                instance_id=instance_id,
                database_id=database_id,
                ddl_statements=ddl_statements,
                task_id="id",
            )
        err = ctx.value
        assert f"The required parameter '{exp_msg}' is empty" in str(err)
        mock_hook.assert_not_called()

    @mock.patch("airflow.providers.google.cloud.operators.spanner.SpannerHook")
    def test_database_update_ex_if_database_not_exist(self, mock_hook):
        mock_hook.return_value.get_database.return_value = None
        op = SpannerUpdateDatabaseInstanceOperator(
            project_id=PROJECT_ID,
            instance_id=INSTANCE_ID,
            database_id=DB_ID,
            ddl_statements=DDL_STATEMENTS,
            task_id="id",
        )
        with pytest.raises(AirflowException) as ctx:
            op.execute(None)
        err = ctx.value
        assert (
            "The Cloud Spanner database 'db1' in project 'project-id' and "
            "instance 'instance-id' is missing" in str(err)
        )
        mock_hook.assert_called_once_with(
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )

    @mock.patch("airflow.providers.google.cloud.operators.spanner.SpannerHook")
    def test_database_delete(self, mock_hook):
        mock_hook.return_value.get_database.return_value = {"name": DB_ID}
        op = SpannerDeleteDatabaseInstanceOperator(
            project_id=PROJECT_ID, instance_id=INSTANCE_ID, database_id=DB_ID, task_id="id"
        )
        result = op.execute(None)
        mock_hook.assert_called_once_with(
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.delete_database.assert_called_once_with(
            project_id=PROJECT_ID, instance_id=INSTANCE_ID, database_id=DB_ID
        )
        assert result

    @mock.patch("airflow.providers.google.cloud.operators.spanner.SpannerHook")
    def test_database_delete_missing_project_id(self, mock_hook):
        mock_hook.return_value.get_database.return_value = {"name": DB_ID}
        op = SpannerDeleteDatabaseInstanceOperator(instance_id=INSTANCE_ID, database_id=DB_ID, task_id="id")
        result = op.execute(None)
        mock_hook.assert_called_once_with(
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.delete_database.assert_called_once_with(
            project_id=None, instance_id=INSTANCE_ID, database_id=DB_ID
        )
        assert result

    @mock.patch("airflow.providers.google.cloud.operators.spanner.SpannerHook")
    def test_database_delete_exits_and_succeeds_if_database_does_not_exist(self, mock_hook):
        mock_hook.return_value.get_database.return_value = None
        op = SpannerDeleteDatabaseInstanceOperator(
            project_id=PROJECT_ID, instance_id=INSTANCE_ID, database_id=DB_ID, task_id="id"
        )
        result = op.execute(None)
        mock_hook.assert_called_once_with(
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.delete_database.assert_not_called()
        assert result

    @pytest.mark.parametrize(
        ("project_id", "instance_id", "database_id", "ddl_statements", "exp_msg"),
        [
            ("", INSTANCE_ID, DB_ID, DDL_STATEMENTS, "project_id"),
            (PROJECT_ID, "", DB_ID, DDL_STATEMENTS, "instance_id"),
            (PROJECT_ID, INSTANCE_ID, "", DDL_STATEMENTS, "database_id"),
        ],
    )
    @mock.patch("airflow.providers.google.cloud.operators.spanner.SpannerHook")
    def test_database_delete_ex_if_param_missing(
        self, mock_hook, project_id, instance_id, database_id, ddl_statements, exp_msg
    ):
        with pytest.raises(AirflowException) as ctx:
            SpannerDeleteDatabaseInstanceOperator(
                project_id=project_id,
                instance_id=instance_id,
                database_id=database_id,
                ddl_statements=ddl_statements,
                task_id="id",
            )
        err = ctx.value
        assert f"The required parameter '{exp_msg}' is empty" in str(err)
        mock_hook.assert_not_called()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest
from google.api_core.exceptions import GoogleAPICallError
from google.cloud.pubsub_v1.types import ReceivedMessage

from airflow.providers.google.cloud.triggers.pubsub import PubsubPullTrigger
from airflow.triggers.base import TriggerEvent

TEST_POLL_INTERVAL = 10
TEST_GCP_CONN_ID = "google_cloud_default"
PROJECT_ID = "test_project_id"
MAX_MESSAGES = 5
ACK_MESSAGES = True


@pytest.fixture
def trigger():
    return PubsubPullTrigger(
        project_id=PROJECT_ID,
        subscription="subscription",
        max_messages=MAX_MESSAGES,
        ack_messages=ACK_MESSAGES,
        poke_interval=TEST_POLL_INTERVAL,
        gcp_conn_id=TEST_GCP_CONN_ID,
        impersonation_chain=None,
    )


async def generate_messages(count: int) -> list[ReceivedMessage]:
    return [
        ReceivedMessage(
            ack_id=f"{i}",
            message={
                "data": f"Message {i}".encode(),
                "attributes": {"type": "generated message"},
            },
        )
        for i in range(1, count + 1)
    ]


class TestPubsubPullTrigger:
    def test_async_pubsub_pull_trigger_serialization_should_execute_successfully(self, trigger):
        """
        Asserts that the PubsubPullTrigger correctly serializes its arguments
        and classpath.
        """
        classpath, kwargs = trigger.serialize()
        assert classpath == "airflow.providers.google.cloud.triggers.pubsub.PubsubPullTrigger"
        assert kwargs == {
            "project_id": PROJECT_ID,
            "subscription": "subscription",
            "max_messages": MAX_MESSAGES,
            "ack_messages": ACK_MESSAGES,
            "poke_interval": TEST_POLL_INTERVAL,
            "gcp_conn_id": TEST_GCP_CONN_ID,
            "impersonation_chain": None,
        }

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.pubsub.PubSubAsyncHook.pull")
    async def test_async_pubsub_pull_trigger_return_event(self, mock_pull):
        mock_pull.return_value = generate_messages(1)
        trigger = PubsubPullTrigger(
            project_id=PROJECT_ID,
            subscription="subscription",
            max_messages=MAX_MESSAGES,
            ack_messages=False,
            poke_interval=TEST_POLL_INTERVAL,
            gcp_conn_id=TEST_GCP_CONN_ID,
            impersonation_chain=None,
        )

        expected_event = TriggerEvent(
            {
                "status": "success",
                "message": [
                    {
                        "ack_id": "1",
                        "message": {
                            "data": "TWVzc2FnZSAx",
                            "attributes": {"type": "generated message"},
                            "message_id": "",
                            "ordering_key": "",
                        },
                        "delivery_attempt": 0,
                    }
                ],
            }
        )

        response = await trigger.run().asend(None)

        assert response == expected_event

    @mock.patch("airflow.providers.google.cloud.triggers.pubsub.PubSubAsyncHook")
    def test_hook(self, mock_async_hook):
        trigger = PubsubPullTrigger(
            project_id=PROJECT_ID,
            subscription="subscription",
            max_messages=MAX_MESSAGES,
            ack_messages=False,
            poke_interval=TEST_POLL_INTERVAL,
            gcp_conn_id=TEST_GCP_CONN_ID,
            impersonation_chain=None,
        )
        async_hook_actual = trigger.hook

        mock_async_hook.assert_called_once_with(
            gcp_conn_id=trigger.gcp_conn_id,
            impersonation_chain=trigger.impersonation_chain,
            project_id=trigger.project_id,
        )
        assert async_hook_actual == mock_async_hook.return_value

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.pubsub.PubSubAsyncHook.pull")
    async def test_async_pubsub_pull_trigger_exception_during_pull(self, mock_pull):
        """Test that exceptions during pull are propagated and not caught."""
        mock_pull.side_effect = GoogleAPICallError("Connection error")

        trigger = PubsubPullTrigger(
            project_id=PROJECT_ID,
            subscription="subscription",
            max_messages=MAX_MESSAGES,
            ack_messages=False,
            poke_interval=TEST_POLL_INTERVAL,
            gcp_conn_id=TEST_GCP_CONN_ID,
            impersonation_chain=None,
        )

        with pytest.raises(GoogleAPICallError, match="Connection error"):
            await trigger.run().asend(None)

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.pubsub.PubSubAsyncHook.acknowledge")
    @mock.patch("airflow.providers.google.cloud.hooks.pubsub.PubSubAsyncHook.pull")
    async def test_async_pubsub_pull_trigger_exception_during_ack(self, mock_pull, mock_acknowledge):
        """Test that exceptions during message acknowledgement are propagated."""
        # Return a coroutine that can be awaited
        mock_pull.return_value = generate_messages(1)
        mock_acknowledge.side_effect = GoogleAPICallError("Acknowledgement failed")

        trigger = PubsubPullTrigger(
            project_id=PROJECT_ID,
            subscription="subscription",
            max_messages=MAX_MESSAGES,
            ack_messages=True,
            poke_interval=TEST_POLL_INTERVAL,
            gcp_conn_id=TEST_GCP_CONN_ID,
            impersonation_chain=None,
        )

        with pytest.raises(GoogleAPICallError, match="Acknowledgement failed"):
            await trigger.run().asend(None)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Module contains a Google Alloy DB Hook."""

from __future__ import annotations

from collections.abc import Sequence
from copy import deepcopy
from typing import TYPE_CHECKING

import tenacity
from google.api_core.gapic_v1.method import DEFAULT, _MethodDefault
from google.cloud import alloydb_v1

from airflow.providers.common.compat.sdk import AirflowException
from airflow.providers.google.common.consts import CLIENT_INFO
from airflow.providers.google.common.hooks.base_google import PROVIDE_PROJECT_ID, GoogleBaseHook

if TYPE_CHECKING:
    import proto
    from google.api_core.operation import Operation
    from google.api_core.retry import Retry
    from google.protobuf.field_mask_pb2 import FieldMask


class AlloyDbHook(GoogleBaseHook):
    """Google Alloy DB Hook."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._client: alloydb_v1.AlloyDBAdminClient | None = None

    def get_alloy_db_admin_client(self) -> alloydb_v1.AlloyDBAdminClient:
        """Retrieve AlloyDB client."""
        if not self._client:
            self._client = alloydb_v1.AlloyDBAdminClient(
                credentials=self.get_credentials(),
                client_info=CLIENT_INFO,
            )
        return self._client

    def wait_for_operation(self, timeout: float | None, operation: Operation) -> proto.Message:
        """Wait for long-lasting operation to complete."""
        self.log.info("Waiting for operation to complete...")
        _timeout: int | None = int(timeout) if timeout else None
        try:
            return operation.result(timeout=_timeout)
        except Exception:
            error = operation.exception(timeout=_timeout)
            raise AirflowException(error)

    @GoogleBaseHook.fallback_to_default_project_id
    def create_cluster(
        self,
        cluster_id: str,
        cluster: alloydb_v1.Cluster | dict,
        location: str,
        project_id: str = PROVIDE_PROJECT_ID,
        request_id: str | None = None,
        validate_only: bool = False,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> Operation:
        """
        Create an Alloy DB cluster.

        .. seealso::
            For more details see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.CreateClusterRequest

        :param cluster_id: Required. ID of the cluster to create.
        :param cluster: Required. Cluster to create. For more details please see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.Cluster
        :param location: Required. The ID of the Google Cloud region where the cluster is located.
        :param project_id: Optional. The ID of the Google Cloud project where the cluster is located.
        :param request_id: Optional. An optional request ID to identify requests. Specify a unique request ID
            so that if you must retry your request, the server ignores the request if it has already been
            completed. The server guarantees that for at least 60 minutes since the first request.
            For example, consider a situation where you make an initial request and the request times out.
            If you make the request again with the same request ID, the server can check if the original operation
            with the same request ID was received, and if so, ignores the second request.
            This prevents clients from accidentally creating duplicate commitments.
            The request ID must be a valid UUID with the exception that zero UUID is not supported
            (00000000-0000-0000-0000-000000000000).
        :param validate_only: Optional. If set, performs request validation, but does not actually execute
            the create request.
        :param retry: Optional. Designation of what errors, if any, should be retried.
        :param timeout: Optional. The timeout for this request.
        :param metadata: Optional. Strings which should be sent along with the request as metadata.
        """
        client = self.get_alloy_db_admin_client()
        return client.create_cluster(
            request={
                "parent": client.common_location_path(project_id, location),
                "cluster_id": cluster_id,
                "cluster": cluster,
                "request_id": request_id,
                "validate_only": validate_only,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

    @GoogleBaseHook.fallback_to_default_project_id
    def create_secondary_cluster(
        self,
        cluster_id: str,
        cluster: alloydb_v1.Cluster | dict,
        location: str,
        project_id: str = PROVIDE_PROJECT_ID,
        request_id: str | None = None,
        validate_only: bool = False,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> Operation:
        """
        Create a secondary Alloy DB cluster.

        .. seealso::
            For more details see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.CreateClusterRequest

        :param cluster_id: Required. ID of the cluster to create.
        :param cluster: Required. Cluster to create. For more details please see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.Cluster
        :param location: Required. The ID of the Google Cloud region where the cluster is located.
        :param project_id: Optional. The ID of the Google Cloud project where the cluster is located.
        :param request_id: Optional. An optional request ID to identify requests. Specify a unique request ID
            so that if you must retry your request, the server ignores the request if it has already been
            completed. The server guarantees that for at least 60 minutes since the first request.
            For example, consider a situation where you make an initial request and the request times out.
            If you make the request again with the same request ID, the server can check if the original operation
            with the same request ID was received, and if so, ignores the second request.
            This prevents clients from accidentally creating duplicate commitments.
            The request ID must be a valid UUID with the exception that zero UUID is not supported
            (00000000-0000-0000-0000-000000000000).
        :param validate_only: Optional. If set, performs request validation, but does not actually execute
            the create request.
        :param retry: Optional. Designation of what errors, if any, should be retried.
        :param timeout: Optional. The timeout for this request.
        :param metadata: Optional. Strings which should be sent along with the request as metadata.
        """
        client = self.get_alloy_db_admin_client()
        return client.create_secondary_cluster(
            request={
                "parent": client.common_location_path(project_id, location),
                "cluster_id": cluster_id,
                "cluster": cluster,
                "request_id": request_id,
                "validate_only": validate_only,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

    @tenacity.retry(
        stop=tenacity.stop_after_attempt(5),
        wait=tenacity.wait_exponential(multiplier=1, max=10),
        retry=tenacity.retry_if_exception_type(ValueError),
    )
    @GoogleBaseHook.fallback_to_default_project_id
    def get_cluster(
        self,
        cluster_id: str,
        location: str,
        project_id: str = PROVIDE_PROJECT_ID,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> alloydb_v1.Cluster:
        """
        Retrieve an Alloy DB cluster.

        .. seealso::
            For more details see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.GetClusterRequest

        :param cluster_id: Required. ID of the cluster.
        :param location: Required. The ID of the Google Cloud region where the cluster is located.
        :param project_id: Optional. The ID of the Google Cloud project where the cluster is located.
        :param retry: Optional. Designation of what errors, if any, should be retried.
        :param timeout: Optional. The timeout for this request.
        :param metadata: Optional. Strings which should be sent along with the request as metadata.
        """
        client = self.get_alloy_db_admin_client()
        return client.get_cluster(
            request={"name": client.cluster_path(project_id, location, cluster_id)},
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

    @GoogleBaseHook.fallback_to_default_project_id
    def update_cluster(
        self,
        cluster_id: str,
        cluster: alloydb_v1.Cluster | dict,
        location: str,
        update_mask: FieldMask | dict | None = None,
        project_id: str = PROVIDE_PROJECT_ID,
        allow_missing: bool = False,
        request_id: str | None = None,
        validate_only: bool = False,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> Operation:
        """
        Update an Alloy DB cluster.

        .. seealso::
            For more details see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.UpdateClusterRequest

        :param cluster_id: Required. ID of the cluster to update.
        :param cluster: Required. Cluster to create. For more details please see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.Cluster
        :param location: Required. The ID of the Google Cloud region where the cluster is located.
        :param update_mask: Optional. Field mask is used to specify the fields to be overwritten in the
            Cluster resource by the update.
        :param request_id: Optional. An optional request ID to identify requests. Specify a unique request ID
            so that if you must retry your request, the server ignores the request if it has already been
            completed. The server guarantees that for at least 60 minutes since the first request.
            For example, consider a situation where you make an initial request and the request times out.
            If you make the request again with the same request ID, the server can check if the original operation
            with the same request ID was received, and if so, ignores the second request.
            This prevents clients from accidentally creating duplicate commitments.
            The request ID must be a valid UUID with the exception that zero UUID is not supported
            (00000000-0000-0000-0000-000000000000).
        :param validate_only: Optional. If set, performs request validation, but does not actually execute
            the create request.
        :param project_id: Optional. The ID of the Google Cloud project where the cluster is located.
        :param allow_missing: Optional. If set to true, update succeeds even if cluster is not found.
            In that case, a new cluster is created and update_mask is ignored.
        :param retry: Optional. Designation of what errors, if any, should be retried.
        :param timeout: Optional. The timeout for this request.
        :param metadata: Optional. Strings which should be sent along with the request as metadata.
        """
        client = self.get_alloy_db_admin_client()
        _cluster = deepcopy(cluster) if isinstance(cluster, dict) else alloydb_v1.Cluster.to_dict(cluster)
        _cluster["name"] = client.cluster_path(project_id, location, cluster_id)
        return client.update_cluster(
            request={
                "update_mask": update_mask,
                "cluster": _cluster,
                "request_id": request_id,
                "validate_only": validate_only,
                "allow_missing": allow_missing,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

    @GoogleBaseHook.fallback_to_default_project_id
    def delete_cluster(
        self,
        cluster_id: str,
        location: str,
        project_id: str = PROVIDE_PROJECT_ID,
        request_id: str | None = None,
        etag: str | None = None,
        validate_only: bool = False,
        force: bool = False,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> Operation:
        """
        Delete an Alloy DB cluster.

        .. seealso::
            For more details see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.DeleteClusterRequest

        :param cluster_id: Required. ID of the cluster to delete.
        :param location: Required. The ID of the Google Cloud region where the cluster is located.
        :param project_id: Optional. The ID of the Google Cloud project where the cluster is located.
        :param request_id: Optional. An optional request ID to identify requests. Specify a unique request ID
            so that if you must retry your request, the server ignores the request if it has already been
            completed. The server guarantees that for at least 60 minutes since the first request.
            For example, consider a situation where you make an initial request and the request times out.
            If you make the request again with the same request ID, the server can check if the original operation
            with the same request ID was received, and if so, ignores the second request.
            This prevents clients from accidentally creating duplicate commitments.
            The request ID must be a valid UUID with the exception that zero UUID is not supported
            (00000000-0000-0000-0000-000000000000).
        :param etag: Optional. The current etag of the Cluster. If an etag is provided and does not match the
            current etag of the Cluster, deletion will be blocked and an ABORTED error will be returned.
        :param validate_only: Optional. If set, performs request validation, but does not actually execute
            the create request.
        :param force: Optional. Whether to cascade delete child instances for given cluster.
        :param retry: Optional. Designation of what errors, if any, should be retried.
        :param timeout: Optional. The timeout for this request.
        :param metadata: Optional. Strings which should be sent along with the request as metadata.
        """
        client = self.get_alloy_db_admin_client()
        return client.delete_cluster(
            request={
                "name": client.cluster_path(project_id, location, cluster_id),
                "request_id": request_id,
                "etag": etag,
                "validate_only": validate_only,
                "force": force,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

    @GoogleBaseHook.fallback_to_default_project_id
    def create_instance(
        self,
        cluster_id: str,
        instance_id: str,
        instance: alloydb_v1.Instance | dict,
        location: str,
        project_id: str = PROVIDE_PROJECT_ID,
        request_id: str | None = None,
        validate_only: bool = False,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> Operation:
        """
        Create an instance in a given Alloy DB cluster.

        .. seealso::
            For more details see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.CreateInstanceRequest

        :param cluster_id: Required. ID of the cluster for creating an instance in.
        :param instance_id: Required. ID of the instance to create.
        :param instance: Required. Instance to create. For more details please see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.Instance
        :param location: Required. The ID of the Google Cloud region where the cluster is located.
        :param project_id: Optional. The ID of the Google Cloud project where the cluster is located.
        :param request_id: Optional. An optional request ID to identify requests. Specify a unique request ID
            so that if you must retry your request, the server ignores the request if it has already been
            completed. The server guarantees that for at least 60 minutes since the first request.
            For example, consider a situation where you make an initial request and the request times out.
            If you make the request again with the same request ID, the server can check if the original operation
            with the same request ID was received, and if so, ignores the second request.
            This prevents clients from accidentally creating duplicate commitments.
            The request ID must be a valid UUID with the exception that zero UUID is not supported
            (00000000-0000-0000-0000-000000000000).
        :param validate_only: Optional. If set, performs request validation, but does not actually execute
            the create request.
        :param retry: Optional. Designation of what errors, if any, should be retried.
        :param timeout: Optional. The timeout for this request.
        :param metadata: Optional. Strings which should be sent along with the request as metadata.
        """
        client = self.get_alloy_db_admin_client()
        return client.create_instance(
            request={
                "parent": client.cluster_path(project_id, location, cluster_id),
                "instance_id": instance_id,
                "instance": instance,
                "request_id": request_id,
                "validate_only": validate_only,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

    @GoogleBaseHook.fallback_to_default_project_id
    def create_secondary_instance(
        self,
        cluster_id: str,
        instance_id: str,
        instance: alloydb_v1.Instance | dict,
        location: str,
        project_id: str = PROVIDE_PROJECT_ID,
        request_id: str | None = None,
        validate_only: bool = False,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> Operation:
        """
        Create a secondary instance in a given Alloy DB cluster.

        .. seealso::
            For more details see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.CreateSecondaryInstanceRequest

        :param cluster_id: Required. ID of the cluster for creating an instance in.
        :param instance_id: Required. ID of the instance to create.
        :param instance: Required. Instance to create. For more details please see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.Instance
        :param location: Required. The ID of the Google Cloud region where the cluster is located.
        :param project_id: Optional. The ID of the Google Cloud project where the cluster is located.
        :param request_id: Optional. An optional request ID to identify requests. Specify a unique request ID
            so that if you must retry your request, the server ignores the request if it has already been
            completed. The server guarantees that for at least 60 minutes since the first request.
            For example, consider a situation where you make an initial request and the request times out.
            If you make the request again with the same request ID, the server can check if the original operation
            with the same request ID was received, and if so, ignores the second request.
            This prevents clients from accidentally creating duplicate commitments.
            The request ID must be a valid UUID with the exception that zero UUID is not supported
            (00000000-0000-0000-0000-000000000000).
        :param validate_only: Optional. If set, performs request validation, but does not actually execute
            the create request.
        :param retry: Optional. Designation of what errors, if any, should be retried.
        :param timeout: Optional. The timeout for this request.
        :param metadata: Optional. Strings which should be sent along with the request as metadata.
        """
        client = self.get_alloy_db_admin_client()
        return client.create_secondary_instance(
            request={
                "parent": client.cluster_path(project_id, location, cluster_id),
                "instance_id": instance_id,
                "instance": instance,
                "request_id": request_id,
                "validate_only": validate_only,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

    @tenacity.retry(
        stop=tenacity.stop_after_attempt(5),
        wait=tenacity.wait_exponential(multiplier=1, max=10),
        retry=tenacity.retry_if_exception_type(ValueError),
    )
    @GoogleBaseHook.fallback_to_default_project_id
    def get_instance(
        self,
        cluster_id: str,
        instance_id: str,
        location: str,
        project_id: str = PROVIDE_PROJECT_ID,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> alloydb_v1.Instance:
        """
        Retrieve an Alloy DB instance.

        .. seealso::
            For more details see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.GetInstanceRequest

        :param cluster_id: Required. ID of the cluster.
        :param instance_id: Required. ID of the instance.
        :param location: Required. The ID of the Google Cloud region where the cluster is located.
        :param project_id: Optional. The ID of the Google Cloud project where the cluster is located.
        :param retry: Optional. Designation of what errors, if any, should be retried.
        :param timeout: Optional. The timeout for this request.
        :param metadata: Optional. Strings which should be sent along with the request as metadata.
        """
        client = self.get_alloy_db_admin_client()
        return client.get_instance(
            request={"name": client.instance_path(project_id, location, cluster_id, instance_id)},
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

    @GoogleBaseHook.fallback_to_default_project_id
    def update_instance(
        self,
        cluster_id: str,
        instance_id: str,
        instance: alloydb_v1.Instance | dict,
        location: str,
        update_mask: FieldMask | dict | None = None,
        project_id: str = PROVIDE_PROJECT_ID,
        allow_missing: bool = False,
        request_id: str | None = None,
        validate_only: bool = False,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> Operation:
        """
        Update an Alloy DB instance.

        .. seealso::
            For more details see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.UpdateInstanceRequest

        :param cluster_id: Required. ID of the cluster.
        :param instance_id: Required. ID of the cluster to update.
        :param instance: Required. Cluster to update. For more details please see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.Instance
        :param location: Required. The ID of the Google Cloud region where the cluster is located.
        :param update_mask: Optional. Field mask is used to specify the fields to be overwritten in the
            Instance resource by the update.
        :param request_id: Optional. The ID of an existing request object.:param request_id: Optional. An optional request ID to identify requests. Specify a unique request ID
            so that if you must retry your request, the server ignores the request if it has already been
            completed. The server guarantees that for at least 60 minutes since the first request.
            For example, consider a situation where you make an initial request and the request times out.
            If you make the request again with the same request ID, the server can check if the original operation
            with the same request ID was received, and if so, ignores the second request.
            This prevents clients from accidentally creating duplicate commitments.
            The request ID must be a valid UUID with the exception that zero UUID is not supported
            (00000000-0000-0000-0000-000000000000).
        :param validate_only: Optional. If set, performs request validation, but does not actually execute
            the create request.
        :param project_id: Optional. The ID of the Google Cloud project where the cluster is located.
        :param allow_missing: Optional. If set to true, update succeeds even if cluster is not found.
            In that case, a new cluster is created and update_mask is ignored.
        :param retry: Optional. Designation of what errors, if any, should be retried.
        :param timeout: Optional. The timeout for this request.
        :param metadata: Optional. Strings which should be sent along with the request as metadata.
        """
        client = self.get_alloy_db_admin_client()
        _instance = (
            deepcopy(instance) if isinstance(instance, dict) else alloydb_v1.Instance.to_dict(instance)
        )
        _instance["name"] = client.instance_path(project_id, location, cluster_id, instance_id)
        return client.update_instance(
            request={
                "update_mask": update_mask,
                "instance": _instance,
                "request_id": request_id,
                "validate_only": validate_only,
                "allow_missing": allow_missing,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

    @GoogleBaseHook.fallback_to_default_project_id
    def delete_instance(
        self,
        instance_id: str,
        cluster_id: str,
        location: str,
        project_id: str = PROVIDE_PROJECT_ID,
        request_id: str | None = None,
        etag: str | None = None,
        validate_only: bool = False,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ):
        """
        Delete an Alloy DB instance.

        .. seealso::
            For more details see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.DeleteInstanceRequest

        :param instance_id: Required. ID of the instance to delete.
        :param cluster_id: Required. ID of the cluster.
        :param location: Required. The ID of the Google Cloud region where the instance is located.
        :param project_id: Optional. The ID of the Google Cloud project where the instance is located.
        :param request_id: Optional. An optional request ID to identify requests. Specify a unique request ID
            so that if you must retry your request, the server ignores the request if it has already been
            completed. The server guarantees that for at least 60 minutes since the first request.
            For example, consider a situation where you make an initial request and the request times out.
            If you make the request again with the same request ID, the server can check if the original operation
            with the same request ID was received, and if so, ignores the second request.
            This prevents clients from accidentally creating duplicate commitments.
            The request ID must be a valid UUID with the exception that zero UUID is not supported
            (00000000-0000-0000-0000-000000000000).
        :param etag: Optional. The current etag of the Instance. If an etag is provided and does not match the
            current etag of the Instance, deletion will be blocked and an ABORTED error will be returned.
        :param validate_only: Optional. If set, performs request validation, but does not actually execute
            the delete request.
        :param retry: Optional. Designation of what errors, if any, should be retried.
        :param timeout: Optional. The timeout for this request.
        :param metadata: Optional. Strings which should be sent along with the request as metadata.
        """
        client = self.get_alloy_db_admin_client()
        return client.delete_instance(
            request={
                "name": client.instance_path(project_id, location, cluster_id, instance_id),
                "request_id": request_id,
                "etag": etag,
                "validate_only": validate_only,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

    @GoogleBaseHook.fallback_to_default_project_id
    def create_user(
        self,
        user_id: str,
        user: alloydb_v1.User | dict,
        cluster_id: str,
        location: str,
        project_id: str = PROVIDE_PROJECT_ID,
        request_id: str | None = None,
        validate_only: bool = False,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> alloydb_v1.User:
        """
        Create a user in a given Alloy DB cluster.

        .. seealso::
            For more details see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.CreateUserRequest

        :param user_id: Required. ID of the user to create.
        :param user: Required. The user to create. For more details please see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.User
        :param cluster_id: Required. ID of the cluster for creating a user in.
        :param location: Required. The ID of the Google Cloud region where the cluster is located.
        :param project_id: Optional. The ID of the Google Cloud project where the cluster is located.
        :param request_id: Optional. An optional request ID to identify requests. Specify a unique request ID
            so that if you must retry your request, the server ignores the request if it has already been
            completed. The server guarantees that for at least 60 minutes since the first request.
            For example, consider a situation where you make an initial request and the request times out.
            If you make the request again with the same request ID, the server can check if the original operation
            with the same request ID was received, and if so, ignores the second request.
            This prevents clients from accidentally creating duplicate commitments.
            The request ID must be a valid UUID with the exception that zero UUID is not supported
            (00000000-0000-0000-0000-000000000000).
        :param validate_only: Optional. If set, performs request validation, but does not actually execute
            the create request.
        :param retry: Optional. Designation of what errors, if any, should be retried.
        :param timeout: Optional. The timeout for this request.
        :param metadata: Optional. Strings which should be sent along with the request as metadata.
        """
        client = self.get_alloy_db_admin_client()
        return client.create_user(
            request={
                "parent": client.cluster_path(project_id, location, cluster_id),
                "user_id": user_id,
                "user": user,
                "request_id": request_id,
                "validate_only": validate_only,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

    @tenacity.retry(
        stop=tenacity.stop_after_attempt(5),
        wait=tenacity.wait_exponential(multiplier=1, max=10),
        retry=tenacity.retry_if_exception_type(ValueError),
    )
    @GoogleBaseHook.fallback_to_default_project_id
    def get_user(
        self,
        user_id: str,
        cluster_id: str,
        location: str,
        project_id: str = PROVIDE_PROJECT_ID,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> alloydb_v1.User:
        """
        Get a user in a given Alloy DB cluster.

        .. seealso::
            For more details see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.GetUserRequest

        :param user_id: Required. ID of the user to create.
        :param cluster_id: Required. ID of the cluster for creating a user in.
        :param location: Required. The ID of the Google Cloud region where the cluster is located.
        :param project_id: Optional. The ID of the Google Cloud project where the cluster is located.
        :param retry: Optional. Designation of what errors, if any, should be retried.
        :param timeout: Optional. The timeout for this request.
        :param metadata: Optional. Strings which should be sent along with the request as metadata.
        """
        client = self.get_alloy_db_admin_client()
        return client.get_user(
            request={
                "name": client.user_path(project_id, location, cluster_id, user_id),
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

    @GoogleBaseHook.fallback_to_default_project_id
    def update_user(
        self,
        cluster_id: str,
        user_id: str,
        user: alloydb_v1.User | dict,
        location: str,
        update_mask: FieldMask | dict | None = None,
        allow_missing: bool = False,
        project_id: str = PROVIDE_PROJECT_ID,
        request_id: str | None = None,
        validate_only: bool = False,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> alloydb_v1.User:
        """
        Update an Alloy DB user.

        .. seealso::
            For more details see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.UpdateUserRequest

        :param cluster_id: Required. ID of the cluster.
        :param user_id: Required. ID of the user to update.
        :param user: Required. User to update. For more details please see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.User
        :param location: Required. The ID of the Google Cloud region where the cluster is located.
        :param update_mask: Optional. Field mask is used to specify the fields to be overwritten in the
            User resource by the update.
        :param request_id: Optional. The ID of an existing request object.:param request_id: Optional. An optional request ID to identify requests. Specify a unique request ID
            so that if you must retry your request, the server ignores the request if it has already been
            completed. The server guarantees that for at least 60 minutes since the first request.
            For example, consider a situation where you make an initial request and the request times out.
            If you make the request again with the same request ID, the server can check if the original operation
            with the same request ID was received, and if so, ignores the second request.
            This prevents clients from accidentally creating duplicate commitments.
            The request ID must be a valid UUID with the exception that zero UUID is not supported
            (00000000-0000-0000-0000-000000000000).
        :param validate_only: Optional. If set, performs request validation, but does not actually execute
            the create request.
        :param allow_missing: Optional. If set to true, update succeeds even if cluster is not found.
            In that case, a new cluster is created and update_mask is ignored.
        :param project_id: Optional. The ID of the Google Cloud project where the cluster is located.
        :param retry: Optional. Designation of what errors, if any, should be retried.
        :param timeout: Optional. The timeout for this request.
        :param metadata: Optional. Strings which should be sent along with the request as metadata.
        """
        client = self.get_alloy_db_admin_client()
        _user = deepcopy(user) if isinstance(user, dict) else alloydb_v1.User.to_dict(user)
        _user["name"] = client.user_path(project_id, location, cluster_id, user_id)
        return client.update_user(
            request={
                "update_mask": update_mask,
                "user": _user,
                "request_id": request_id,
                "validate_only": validate_only,
                "allow_missing": allow_missing,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

    @GoogleBaseHook.fallback_to_default_project_id
    def delete_user(
        self,
        user_id: str,
        cluster_id: str,
        location: str,
        project_id: str = PROVIDE_PROJECT_ID,
        request_id: str | None = None,
        validate_only: bool = False,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ):
        """
        Delete an Alloy DB user.

        .. seealso::
            For more details see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.DeleteUserRequest

        :param user_id: Required. ID of the user to delete.
        :param cluster_id: Required. ID of the cluster.
        :param location: Required. The ID of the Google Cloud region where the instance is located.
        :param project_id: Optional. The ID of the Google Cloud project where the instance is located.
        :param request_id: Optional. An optional request ID to identify requests. Specify a unique request ID
            so that if you must retry your request, the server ignores the request if it has already been
            completed. The server guarantees that for at least 60 minutes since the first request.
            For example, consider a situation where you make an initial request and the request times out.
            If you make the request again with the same request ID, the server can check if the original operation
            with the same request ID was received, and if so, ignores the second request.
            This prevents clients from accidentally creating duplicate commitments.
            The request ID must be a valid UUID with the exception that zero UUID is not supported
            (00000000-0000-0000-0000-000000000000).
        :param validate_only: Optional. If set, performs request validation, but does not actually execute
            the delete request.
        :param retry: Optional. Designation of what errors, if any, should be retried.
        :param timeout: Optional. The timeout for this request.
        :param metadata: Optional. Strings which should be sent along with the request as metadata.
        """
        client = self.get_alloy_db_admin_client()
        return client.delete_user(
            request={
                "name": client.user_path(project_id, location, cluster_id, user_id),
                "request_id": request_id,
                "validate_only": validate_only,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

    @GoogleBaseHook.fallback_to_default_project_id
    def create_backup(
        self,
        backup_id: str,
        backup: alloydb_v1.Backup | dict,
        location: str,
        project_id: str = PROVIDE_PROJECT_ID,
        request_id: str | None = None,
        validate_only: bool = False,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> Operation:
        """
        Create a backup in a given Alloy DB cluster.

        .. seealso::
            For more details see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.CreateBackupRequest

        :param backup_id: Required. ID of the backup to create.
        :param backup: Required. The backup to create. For more details please see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.Backup
        :param location: Required. The ID of the Google Cloud region where the cluster is located.
        :param project_id: Optional. The ID of the Google Cloud project where the cluster is located.
        :param request_id: Optional. An optional request ID to identify requests. Specify a unique request ID
            so that if you must retry your request, the server ignores the request if it has already been
            completed. The server guarantees that for at least 60 minutes since the first request.
            For example, consider a situation where you make an initial request and the request times out.
            If you make the request again with the same request ID, the server can check if the original operation
            with the same request ID was received, and if so, ignores the second request.
            This prevents clients from accidentally creating duplicate commitments.
            The request ID must be a valid UUID with the exception that zero UUID is not supported
            (00000000-0000-0000-0000-000000000000).
        :param validate_only: Optional. If set, performs request validation, but does not actually execute
            the create request.
        :param retry: Optional. Designation of what errors, if any, should be retried.
        :param timeout: Optional. The timeout for this request.
        :param metadata: Optional. Strings which should be sent along with the request as metadata.
        """
        client = self.get_alloy_db_admin_client()
        return client.create_backup(
            request={
                "parent": client.common_location_path(project_id, location),
                "backup_id": backup_id,
                "backup": backup,
                "request_id": request_id,
                "validate_only": validate_only,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

    @tenacity.retry(
        stop=tenacity.stop_after_attempt(5),
        wait=tenacity.wait_exponential(multiplier=1, max=10),
        retry=tenacity.retry_if_exception_type(ValueError),
    )
    @GoogleBaseHook.fallback_to_default_project_id
    def get_backup(
        self,
        backup_id: str,
        location: str,
        project_id: str = PROVIDE_PROJECT_ID,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> alloydb_v1.Backup:
        """
        Get a backup in a given Alloy DB cluster.

        .. seealso::
            For more details see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.GetBackupRequest

        :param backup_id: Required. ID of the backup to create.
        :param location: Required. The ID of the Google Cloud region where the cluster is located.
        :param project_id: Optional. The ID of the Google Cloud project where the cluster is located.
        :param retry: Optional. Designation of what errors, if any, should be retried.
        :param timeout: Optional. The timeout for this request.
        :param metadata: Optional. Strings which should be sent along with the request as metadata.
        """
        client = self.get_alloy_db_admin_client()
        return client.get_backup(
            request={
                "name": client.backup_path(project_id, location, backup_id),
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

    @GoogleBaseHook.fallback_to_default_project_id
    def update_backup(
        self,
        backup_id: str,
        backup: alloydb_v1.Backup | dict,
        location: str,
        update_mask: FieldMask | dict | None = None,
        allow_missing: bool = False,
        project_id: str = PROVIDE_PROJECT_ID,
        request_id: str | None = None,
        validate_only: bool = False,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> Operation:
        """
        Update an Alloy DB backup.

        .. seealso::
            For more details see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.UpdateBackupRequest

        :param backup_id: Required. ID of the backup to update.
        :param backup: Required. Backup to update. For more details please see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.Backup
        :param location: Required. The ID of the Google Cloud region where the cluster is located.
        :param update_mask: Optional. Field mask is used to specify the fields to be overwritten in the
            Backup resource by the update.
        :param request_id: Optional. The ID of an existing request object.:param request_id: Optional. An optional request ID to identify requests. Specify a unique request ID
            so that if you must retry your request, the server ignores the request if it has already been
            completed. The server guarantees that for at least 60 minutes since the first request.
            For example, consider a situation where you make an initial request and the request times out.
            If you make the request again with the same request ID, the server can check if the original operation
            with the same request ID was received, and if so, ignores the second request.
            This prevents clients from accidentally creating duplicate commitments.
            The request ID must be a valid UUID with the exception that zero UUID is not supported
            (00000000-0000-0000-0000-000000000000).
        :param validate_only: Optional. If set, performs request validation, but does not actually execute
            the create request.
        :param allow_missing: Optional. If set to true, update succeeds even if cluster is not found.
            In that case, a new cluster is created and update_mask is ignored.
        :param project_id: Optional. The ID of the Google Cloud project where the cluster is located.
        :param retry: Optional. Designation of what errors, if any, should be retried.
        :param timeout: Optional. The timeout for this request.
        :param metadata: Optional. Strings which should be sent along with the request as metadata.
        """
        client = self.get_alloy_db_admin_client()
        _backup = deepcopy(backup) if isinstance(backup, dict) else alloydb_v1.Backup.to_dict(backup)
        _backup["name"] = client.backup_path(project_id, location, backup_id)
        return client.update_backup(
            request={
                "update_mask": update_mask,
                "backup": _backup,
                "request_id": request_id,
                "validate_only": validate_only,
                "allow_missing": allow_missing,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

    @GoogleBaseHook.fallback_to_default_project_id
    def delete_backup(
        self,
        backup_id: str,
        location: str,
        project_id: str = PROVIDE_PROJECT_ID,
        request_id: str | None = None,
        validate_only: bool = False,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ):
        """
        Delete an Alloy DB backup.

        .. seealso::
            For more details see API documentation:
            https://cloud.google.com/python/docs/reference/alloydb/latest/google.cloud.alloydb_v1.types.DeleteBackupRequest

        :param backup_id: Required. ID of the backup to delete.
        :param location: Required. The ID of the Google Cloud region where the instance is located.
        :param project_id: Optional. The ID of the Google Cloud project where the instance is located.
        :param request_id: Optional. An optional request ID to identify requests. Specify a unique request ID
            so that if you must retry your request, the server ignores the request if it has already been
            completed. The server guarantees that for at least 60 minutes since the first request.
            For example, consider a situation where you make an initial request and the request times out.
            If you make the request again with the same request ID, the server can check if the original operation
            with the same request ID was received, and if so, ignores the second request.
            This prevents clients from accidentally creating duplicate commitments.
            The request ID must be a valid UUID with the exception that zero UUID is not supported
            (00000000-0000-0000-0000-000000000000).
        :param validate_only: Optional. If set, performs request validation, but does not actually execute
            the delete request.
        :param retry: Optional. Designation of what errors, if any, should be retried.
        :param timeout: Optional. The timeout for this request.
        :param metadata: Optional. Strings which should be sent along with the request as metadata.
        """
        client = self.get_alloy_db_admin_client()
        return client.delete_backup(
            request={
                "name": client.backup_path(project_id, location, backup_id),
                "request_id": request_id,
                "validate_only": validate_only,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains Google Stackdriver links."""

from __future__ import annotations

from airflow.providers.google.cloud.links.base import BaseGoogleLink

STACKDRIVER_BASE_LINK = "/monitoring/alerting"
STACKDRIVER_NOTIFICATIONS_LINK = STACKDRIVER_BASE_LINK + "/notifications?project={project_id}"
STACKDRIVER_POLICIES_LINK = STACKDRIVER_BASE_LINK + "/policies?project={project_id}"


class StackdriverNotificationsLink(BaseGoogleLink):
    """Helper class for constructing Stackdriver Notifications Link."""

    name = "Cloud Monitoring Notifications"
    key = "stackdriver_notifications"
    format_str = STACKDRIVER_NOTIFICATIONS_LINK


class StackdriverPoliciesLink(BaseGoogleLink):
    """Helper class for constructing Stackdriver Policies Link."""

    name = "Cloud Monitoring Policies"
    key = "stackdriver_policies"
    format_str = STACKDRIVER_POLICIES_LINK
